import json
import os
import shutil
from random import choice
from time import time

import cv2

from maps4fs import DTMProvider, Map
from maps4fs.generator.game import Game
from maps4fs.generator.settings import (
    BackgroundSettings,
    DEMSettings,
    SatelliteSettings,
)

working_directory = os.getcwd()

base_directory = os.path.join(working_directory, "tests/data")
if os.path.isdir(base_directory):
    shutil.rmtree(base_directory)

coordinates_cases = [
    (45.369648574398184, 19.801106980618925),
    (39.51147544442993, -115.96064194571787),
    (-13.038004302866275, -57.09179831840436),
    (-15.912277514425883, 25.9557832265989),
    (-35.95760563718185, 149.1495358824173),
    (58.52085065306593, 31.27771396353221),
    (35.25541295723034, 139.04857855524995),
]

game_code_cases = ["FS25"]

dtm_provider_code = "srtm30"
dtm_provider = DTMProvider.get_provider_by_code(dtm_provider_code)


background_settings = BackgroundSettings(
    generate_background=True,
    generate_water=True,
    resize_factor=16,
    remove_center=False,
)


def get_random_size() -> tuple[int, int]:
    """Return random size.

    Returns:
        tuple[int, int]: Random size.
    """
    sizes_cases = [(1024, 1024), (2048, 2048), (4096, 4096), (8192, 8192)]
    return choice(sizes_cases[:2])  # Larger maps are too slow for automated tests.


def map_directory(game_code: str) -> str:
    """Creates a new map directory and returns its path.

    Arguments:
        game_code {str} -- Game code.

    Returns:
        str: Path to the new map directory.
    """
    timestamp = int(time())
    directory = os.path.join(base_directory, f"{game_code}_{timestamp}")
    os.makedirs(directory, exist_ok=True)
    return directory


def load_textures_schema(json_path: str) -> dict:
    """Load textures schema from JSON file.

    Arguments:
        json_path (str): Path to the JSON file.

    Returns:
        dict: Loaded JSON file.
    """
    with open(json_path, "r") as file:
        return json.load(file)


def test_map():
    """Test Map generation for different cases."""
    for game_code in game_code_cases:
        game = Game.from_code(game_code)
        for coordinates in coordinates_cases:
            height, width = get_random_size()
            directory = map_directory(game_code)

            print(f"Generating map for {game_code} at {coordinates} with size {height}x{width}...")

            map = Map(
                game=game,
                dtm_provider=dtm_provider,
                dtm_provider_settings=None,
                background_settings=background_settings,
                coordinates=coordinates,
                size=height,
                rotation=0,
                map_directory=directory,
            )

            for _ in map.generate():
                pass

            layers_schema = load_textures_schema(game.texture_schema)

            texture_subdir = "maps/map/data" if game_code == "FS22" else "map/data"

            textures_directory = os.path.join(directory, texture_subdir)
            for texture in layers_schema:
                texture_name = texture["name"]
                numer_of_layers = texture["count"]

                exclude_weight = texture.get("exclude_weight", False)
                if exclude_weight:
                    continue

                if numer_of_layers == 0:
                    continue
                for idx in range(1, numer_of_layers + 1):
                    texture_path = os.path.join(
                        textures_directory, f"{texture_name}{str(idx).zfill(2)}_weight.png"
                    )
                    assert os.path.isfile(texture_path), f"Texture not found: {texture_path}"
                    img = cv2.imread(texture_path)
                    assert img is not None, f"Texture could not be read: {texture_path}"
                    assert img.shape == (
                        height,
                        width,
                        3,
                    ), f"Texture shape mismatch: {img.shape} != {(height, width, 3)}"
                    assert img.dtype == "uint8", f"Texture dtype mismatch: {img.dtype} != uint8"

            dem_name = "map_dem.png" if game_code == "FS22" else "dem.png"

            dem_file = os.path.join(textures_directory, dem_name)
            assert os.path.isfile(dem_file), f"DEM file not found: {dem_file}"
            img = cv2.imread(dem_file, cv2.IMREAD_UNCHANGED)
            assert img is not None, f"DEM could not be read: {dem_file}"

            assert img.dtype == "uint16", f"DEM dtype mismatch: {img.dtype} != uint16"


def test_map_preview():
    """Test Map preview generation."""
    case = choice(coordinates_cases)

    game_code = choice(game_code_cases)
    game = Game.from_code(game_code)

    height, width = get_random_size()

    directory = map_directory(game_code)
    map = Map(
        game=game,
        dtm_provider=dtm_provider,
        dtm_provider_settings=None,
        background_settings=background_settings,
        coordinates=case,
        size=height,
        rotation=0,
        map_directory=directory,
    )
    for _ in map.generate():
        pass
    previews_paths = map.previews()
    for preview_path in previews_paths:
        if not preview_path.endswith(".png"):
            continue
        assert os.path.isfile(preview_path), f"Preview not found: {preview_path}"
        img = cv2.imread(preview_path)
        assert img is not None, f"Preview could not be read: {preview_path}"


def test_map_pack():
    """Test Map packing into zip archive."""
    case = choice(coordinates_cases)

    game_code = choice(game_code_cases)
    game = Game.from_code(game_code)

    height, width = get_random_size()

    dem_settings = DEMSettings(multiplier=2, blur_radius=15, plateau=1000, water_depth=500)

    satellite_settings = SatelliteSettings(download_images=True, satellite_margin=0, zoom_level=14)

    directory = map_directory(game_code)
    map = Map(
        game=game,
        dtm_provider=dtm_provider,
        dtm_provider_settings=None,
        coordinates=case,
        size=height,
        rotation=30,
        map_directory=directory,
        dem_settings=dem_settings,
        background_settings=background_settings,
        satellite_settings=satellite_settings,
    )
    for _ in map.generate():
        pass
    archive_name = os.path.join(base_directory, "archive")
    archive_path = map.pack(archive_name)
    assert os.path.isfile(archive_path), f"Archive not found: {archive_path}"
    unpacked_directory = os.path.join(base_directory, "unpacked")
    os.makedirs(unpacked_directory, exist_ok=True)
    try:
        shutil.unpack_archive(archive_path, unpacked_directory)
    except Exception as e:
        assert False, f"Archive could not be unpacked: {e}"
    assert os.path.isdir(unpacked_directory), f"Unpacked directory not found: {unpacked_directory}"



"""This module contains provider of Sachsen-Anhalt data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class SachsenAnhaltProvider(WCSProvider, DTMProvider):
    """Provider of Sachsen-Anhalt data."""

    _code = "sachsen-anhalt"
    _name = "Sachsen-Anhalt"
    _region = "DE"
    _icon = "🇩🇪"
    _resolution = 1
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _is_base = False
    _extents = [
        (53.0769416826493412, 50.8927195980075453, 13.3232545527125836, 10.5092298520646867)
    ]

    _url = "https://www.geodatenportal.sachsen-anhalt.de/wss/service/ST_LVermGeo_DGM1_WCS_OpenData/guest"  # pylint: disable=line-too-long
    _wcs_version = "1.0.0"
    _source_crs = "EPSG:4326"
    _tile_size = 0.02

    def get_wcs_parameters(self, tile: tuple[float, float, float, float]) -> dict:
        return {
            "identifier": "1",
            "bbox": tile,
            "format": "GeoTIFF",
            "crs": "EPSG:4326",
            "width": 1000,
            "height": 1000,
            "timeout": 600,
        }



"""This module contains the class representing a layer with textures and tags."""

from __future__ import annotations

import os
import re


class Layer:
    """Class which represents a layer with textures and tags.
    It's using to obtain data from OSM using tags and make changes into corresponding textures.

    Arguments:
        name (str): Name of the layer.
        tags (dict[str, str | list[str]]): Dictionary of tags to search for.
        width (int | None): Width of the polygon in meters (only for LineString).
        color (tuple[int, int, int]): Color of the layer in BGR format.
        exclude_weight (bool): Flag to exclude weight from the texture.
        priority (int | None): Priority of the layer.
        info_layer (str | None): Name of the corresnponding info layer.
        usage (str | None): Usage of the layer.
        background (bool): Flag to determine if the layer is a background.
        invisible (bool): Flag to determine if the layer is invisible.

    Attributes:
        name (str): Name of the layer.
        tags (dict[str, str | list[str]]): Dictionary of tags to search for.
        width (int | None): Width of the polygon in meters (only for LineString).
    """

    # pylint: disable=R0913
    def __init__(  # pylint: disable=R0917
        self,
        name: str,
        count: int,
        tags: dict[str, str | list[str] | bool] | None = None,
        width: int | None = None,
        color: tuple[int, int, int] | list[int] | None = None,
        exclude_weight: bool = False,
        priority: int | None = None,
        info_layer: str | None = None,
        usage: str | None = None,
        background: bool = False,
        invisible: bool = False,
        procedural: list[str] | None = None,
        border: int | None = None,
        precise_tags: dict[str, str | list[str] | bool] | None = None,
        precise_usage: str | None = None,
    ):
        self.name = name
        self.count = count
        self.tags = tags
        self.width = width
        self.color = color if color else (255, 255, 255)
        self.exclude_weight = exclude_weight
        self.priority = priority
        self.info_layer = info_layer
        self.usage = usage
        self.background = background
        self.invisible = invisible
        self.procedural = procedural
        self.border = border
        self.precise_tags = precise_tags
        self.precise_usage = precise_usage

    def to_json(self) -> dict[str, str | list[str] | bool]:  # type: ignore
        """Returns dictionary with layer data.

        Returns:
            dict: Dictionary with layer data."""
        data = {
            "name": self.name,
            "count": self.count,
            "tags": self.tags,
            "width": self.width,
            "color": list(self.color),
            "exclude_weight": self.exclude_weight,
            "priority": self.priority,
            "info_layer": self.info_layer,
            "usage": self.usage,
            "background": self.background,
            "invisible": self.invisible,
            "procedural": self.procedural,
            "border": self.border,
            "precise_tags": self.precise_tags,
            "precise_usage": self.precise_usage,
        }

        data = {k: v for k, v in data.items() if v is not None}
        return data  # type: ignore

    @classmethod
    def from_json(cls, data: dict[str, str | list[str] | bool]) -> Layer:
        """Creates a new instance of the class from dictionary.

        Arguments:
            data (dict[str, str | list[str] | bool]): Dictionary with layer data.

        Returns:
            Layer: New instance of the class.
        """
        return cls(**data)  # type: ignore

    def path(self, weights_directory: str) -> str:
        """Returns path to the first texture of the layer.

        Arguments:
            weights_directory (str): Path to the directory with weights.

        Returns:
            str: Path to the texture.
        """
        idx = "01" if self.count > 0 else ""
        weight_postfix = "_weight" if not self.exclude_weight else ""
        return os.path.join(weights_directory, f"{self.name}{idx}{weight_postfix}.png")

    def path_preview(self, weights_directory: str) -> str:
        """Returns path to the preview of the first texture of the layer.

        Arguments:
            weights_directory (str): Path to the directory with weights.

        Returns:
            str: Path to the preview.
        """
        return self.path(weights_directory).replace(".png", "_preview.png")

    def get_preview_or_path(self, weights_directory: str) -> str:
        """Returns path to the preview of the first texture of the layer if it exists,
        otherwise returns path to the texture.

        Arguments:
            weights_directory (str): Path to the directory with weights.

        Returns:
            str: Path to the preview or texture.
        """
        preview_path = self.path_preview(weights_directory)
        return preview_path if os.path.isfile(preview_path) else self.path(weights_directory)

    def paths(self, weights_directory: str) -> list[str]:
        """Returns a list of paths to the textures of the layer.
        NOTE: Works only after the textures are generated, since it just lists the directory.

        Arguments:
            weights_directory (str): Path to the directory with weights.

        Returns:
            list[str]: List of paths to the textures.
        """
        weight_files = os.listdir(weights_directory)

        # Inconsistent names are the name of textures that are not following the pattern
        # of texture_name{idx}_weight.png.
        inconsistent_names = ["forestRockRoot", "waterPuddle"]

        if self.name in inconsistent_names:
            return [
                os.path.join(weights_directory, weight_file)
                for weight_file in weight_files
                if weight_file.startswith(self.name)
            ]

        return [
            os.path.join(weights_directory, weight_file)
            for weight_file in weight_files
            if re.match(rf"{self.name}\d{{2}}_weight.png", weight_file)
        ]



"""This module contains DEM class for processing Digital Elevation Model data."""

import math
from typing import Any

import cv2
import numpy as np

# import rasterio  # type: ignore
from pympler import asizeof  # type: ignore

from maps4fs.generator.component.base.component import Component
from maps4fs.generator.dtm.dtm import DTMProvider


# pylint: disable=R0903, R0902
class DEM(Component):
    """Component for processing Digital Elevation Model data.

    Arguments:
        game (Game): The game instance for which the map is generated.
        coordinates (tuple[float, float]): The latitude and longitude of the center of the map.
        map_size (int): The size of the map in pixels.
        map_rotated_size (int): The size of the map in pixels after rotation.
        rotation (int): The rotation angle of the map.
        map_directory (str): The directory where the map files are stored.
        logger (Any, optional): The logger to use. Must have at least three basic methods: debug,
            info, warning. If not provided, default logging will be used.
    """

    def preprocess(self) -> None:
        self._dem_path = self.game.dem_file_path(self.map_directory)
        self.temp_dir = "temp"

        self.logger.debug("Map size: %s x %s.", self.map_size, self.map_size)
        self.logger.debug(
            "Map rotated size: %s x %s.", self.map_rotated_size, self.map_rotated_size
        )

        self.output_resolution = self.get_output_resolution()
        self.logger.debug("Output resolution for DEM data: %s.", self.output_resolution)

        blur_radius = self.map.dem_settings.blur_radius
        if blur_radius is None or blur_radius <= 0:
            # We'll disable blur if the radius is 0 or negative.
            blur_radius = 0
        elif blur_radius % 2 == 0:
            blur_radius += 1
        self.blur_radius = blur_radius
        self.multiplier = self.map.dem_settings.multiplier
        self.logger.debug(
            "DEM value multiplier is %s, blur radius is %s.",
            self.multiplier,
            self.blur_radius,
        )

        self.dtm_provider: DTMProvider = self.map.dtm_provider(  # type: ignore
            coordinates=self.coordinates,
            user_settings=self.map.dtm_provider_settings,
            size=self.map_rotated_size,
            directory=self.temp_dir,
            logger=self.logger,
            map=self.map,
        )

        self.info: dict[str, Any] = {}

    @property
    def dem_path(self) -> str:
        """Returns path to the DEM file.

        Returns:
            str: Path to the DEM file.
        """
        return self._dem_path

    # pylint: disable=W0201
    def set_dem_path(self, dem_path: str) -> None:
        """Set path to the DEM file.

        Arguments:
            dem_path (str): Path to the DEM file.
        """
        self._dem_path = dem_path

    # pylint: disable=W0201
    def set_output_resolution(self, output_resolution: tuple[int, int]) -> None:
        """Set output resolution for DEM data (width, height).

        Arguments:
            output_resolution (tuple[int, int]): Output resolution for DEM data.
        """
        self.output_resolution = output_resolution

    def get_output_resolution(self, use_original: bool = False) -> tuple[int, int]:
        """Get output resolution for DEM data.

        Arguments:
            use_original (bool, optional): If True, will use original map size. Defaults to False.

        Returns:
            tuple[int, int]: Output resolution for DEM data.
        """
        map_size = self.map_size if use_original else self.map_rotated_size

        dem_size = int((map_size / 2) * self.game.dem_multipliyer)

        self.logger.debug(
            "DEM size multiplier is %s, DEM size: %sx%s, use original: %s.",
            self.game.dem_multipliyer,
            dem_size,
            dem_size,
            use_original,
        )
        return dem_size, dem_size

    def update_info(self, state: str, data: np.ndarray) -> None:
        """Update info dictionary with additional information about DEM data.

        Arguments:
            data (np.ndarray): DEM data.
        """

        try:
            entry = {
                "min": float(data.min()),
                "max": float(data.max()),
                "deviation": float(data.max() - data.min()),
                "dtype": str(data.dtype),
                "shape": str(data.shape),
            }
            self.info[state] = entry
        except Exception as e:
            self.logger.warning("Failed to update DEM info: %s.", e)

    def process(self) -> None:
        """Reads DTM file, crops it to map size, normalizes and blurs it,
        saves to map directory."""

        dem_output_resolution = self.output_resolution
        self.logger.debug("DEM output resolution: %s.", dem_output_resolution)

        try:
            data = self.dtm_provider.get_numpy()
        except Exception as e:  # pylint: disable=W0718
            self.logger.error("Failed to get DEM data from DTM provider: %s.", e)
            raise ValueError(
                f"Failed to get DEM data from DTM provider: {e}. "
                "Try using different DTM provider."
            )

        if len(data.shape) != 2:
            self.logger.error("DTM provider returned incorrect data: more than 1 channel.")
            raise ValueError(
                "DTM provider returned incorrect data: more than 1 channel. "
                "Try using different DTM provider."
            )

        if data.dtype not in ["int16", "uint16", "float", "float32"]:
            self.logger.error("DTM provider returned incorrect data type: %s.", data.dtype)
            raise ValueError(
                f"DTM provider returned incorrect data type: {data.dtype}. "
                "Try using different DTM provider."
            )

        self.logger.debug(
            "DEM data was retrieved from DTM provider. Shape: %s, dtype: %s. Min: %s, max: %s.",
            data.shape,
            data.dtype,
            data.min(),
            data.max(),
        )

        self.update_info("original", data)

        # Check if the data contains any non zero values, otherwise raise an error.
        if not np.any(data):
            self.logger.error("DTM provider returned empty data.")
            raise ValueError("DTM provider returned empty data. Try using different DTM provider.")

        # 1. Resize DEM data to the output resolution.
        resampled_data = self.resize_to_output(data)

        # 2. Apply multiplier (-10 to 120.4 becomes -20 to 240.8)
        resampled_data = self.apply_multiplier(resampled_data)
        self.update_info("multiplied", resampled_data)

        # 3. Raise terrain, and optionally lower to plateau level+water depth
        # e.g. -20 to 240.8m becomes 20 to 280.8m
        resampled_data = self.raise_or_lower(resampled_data)
        self.update_info("raised_lowered", resampled_data)

        # 4. Determine actual height scale value using ceiling
        # e.g. 255 becomes 280.8+10 = 291
        height_scale_value = self.determine_height_scale(resampled_data)

        # 5. Normalize DEM data to 16-bit unsigned integer range (0 to 65535)
        # e.g. multiply by 65535/291, clip and return as uint16
        resampled_data = self.normalize_data(resampled_data, height_scale_value)
        self.update_info("normalized", resampled_data)

        # 6. Blur DEM data.
        resampled_data = self.apply_blur(resampled_data)

        cv2.imwrite(self._dem_path, resampled_data)
        self.logger.debug("DEM data was saved to %s.", self._dem_path)

        if self.rotation:
            self.rotate_dem()

    def normalize_data(self, data: np.ndarray, height_scale_value: int) -> np.ndarray:
        """Normalize DEM data to 16-bit unsigned integer range (0 to 65535).

        Arguments:
            data (np.ndarray): DEM data.
            height_scale_value (int): Height scale value.

        Returns:
            np.ndarray: Normalized DEM data.
        """
        normalized_data = np.clip((data / height_scale_value) * 65535, 0, 65535).astype(np.uint16)
        self.logger.debug(
            "DEM data was normalized and clipped to 16-bit unsigned integer range. "
            "Min: %s, max: %s.",
            normalized_data.min(),
            normalized_data.max(),
        )
        return normalized_data

    def determine_height_scale(self, data: np.ndarray) -> int:
        """Determine height scale value using ceiling.

        Arguments:
            data (np.ndarray): DEM data.

        Returns:
            int: Height scale value.
        """
        height_scale = self.map.dem_settings.minimum_height_scale
        adjusted_height_scale = math.ceil(
            max(height_scale, data.max() + self.map.dem_settings.ceiling)
        )

        self.map.shared_settings.height_scale_value = adjusted_height_scale  # type: ignore
        self.map.shared_settings.mesh_z_scaling_factor = 65535 / adjusted_height_scale
        self.map.shared_settings.height_scale_multiplier = adjusted_height_scale / 255
        self.map.shared_settings.change_height_scale = True  # type: ignore

        try:
            entry = {
                "height_scale_from_settings": height_scale,
                "adjusted_height_scale": adjusted_height_scale,
                "mesh_z_scaling_factor": self.map.shared_settings.mesh_z_scaling_factor,
                "height_scale_multiplier": self.map.shared_settings.height_scale_multiplier,
            }
            self.info["height_scale"] = entry
        except Exception as e:
            self.logger.warning("Failed to update DEM info: %s.", e)

        self.logger.debug("Height scale value is %s.", adjusted_height_scale)
        return adjusted_height_scale

    def raise_or_lower(self, data: np.ndarray) -> np.ndarray:
        """Raise or lower terrain to the level of plateau+water depth."""

        if not self.map.dem_settings.adjust_terrain_to_ground_level:
            return data

        desired_ground_level = self.map.dem_settings.plateau + self.map.dem_settings.water_depth
        current_ground_level = data.min()

        data = data + (desired_ground_level - current_ground_level)

        self.logger.debug(
            "Array was shifted to the ground level %s. Min: %s, max: %s.",
            desired_ground_level,
            data.min(),
            data.max(),
        )
        return data

    def apply_multiplier(self, data: np.ndarray) -> np.ndarray:
        """Apply multiplier to DEM data.

        Arguments:
            data (np.ndarray): DEM data.

        Returns:
            np.ndarray: Multiplied DEM data.
        """
        if not self.multiplier != 1:
            return data

        multiplied_data = data * self.multiplier
        self.logger.debug(
            "DEM data was multiplied by %s. Min: %s, max: %s.",
            self.multiplier,
            multiplied_data.min(),
            multiplied_data.max(),
        )
        return multiplied_data

    def resize_to_output(self, data: np.ndarray) -> np.ndarray:
        """Resize DEM data to the output resolution.

        Arguments:
            data (np.ndarray): DEM data.

        Returns:
            np.ndarray: Resized DEM data.
        """
        resampled_data = cv2.resize(data, self.output_resolution, interpolation=cv2.INTER_LINEAR)

        size_of_resampled_data = asizeof.asizeof(resampled_data) / 1024 / 1024
        self.logger.debug("Size of resampled data: %s MB.", size_of_resampled_data)

        return resampled_data

    def apply_blur(self, data: np.ndarray) -> np.ndarray:
        """Apply blur to DEM data.

        Arguments:
            data (np.ndarray): DEM data.

        Returns:
            np.ndarray: Blurred DEM data.
        """
        if self.blur_radius == 0:
            return data

        self.logger.debug(
            "Applying Gaussion blur to DEM data with kernel size %s.",
            self.blur_radius,
        )

        blurred_data = cv2.GaussianBlur(
            data, (self.blur_radius, self.blur_radius), sigmaX=10, sigmaY=10
        )
        self.logger.debug(
            "DEM data was blurred. Shape: %s, dtype: %s. Min: %s, max: %s.",
            blurred_data.shape,
            blurred_data.dtype,
            blurred_data.min(),
            blurred_data.max(),
        )
        return blurred_data

    def rotate_dem(self) -> None:
        """Rotate DEM image."""
        self.logger.debug("Rotating DEM image by %s degrees.", self.rotation)
        output_width, output_height = self.get_output_resolution(use_original=True)

        self.logger.debug(
            "Output resolution for rotated DEM: %s x %s.", output_width, output_height
        )

        self.rotate_image(
            self._dem_path,
            self.rotation,
            output_height=output_height,
            output_width=output_width,
        )

    def info_sequence(self) -> dict[Any, Any] | None:  # type: ignore
        """Returns the information sequence for the component.

        Returns:
            dict[Any, Any]: The information sequence for the component.
        """
        return self.info



"""This module contains Map class, which is used to generate map using all components."""

from __future__ import annotations

import json
import os
import shutil
from typing import Any, Generator

from geopy.geocoders import Nominatim

from maps4fs.generator.component import Background, Component, Layer, Texture
from maps4fs.generator.dtm.dtm import DTMProvider, DTMProviderSettings
from maps4fs.generator.game import Game
from maps4fs.generator.settings import (
    BackgroundSettings,
    DEMSettings,
    GRLESettings,
    I3DSettings,
    SatelliteSettings,
    SharedSettings,
    SplineSettings,
    TextureSettings,
)
from maps4fs.generator.statistics import send_advanced_settings, send_main_settings
from maps4fs.logger import Logger


class Map:
    """Class used to generate map using all components.

    Arguments:
        game (Type[Game]): Game for which the map is generated.
        coordinates (tuple[float, float]): Coordinates of the center of the map.
        size (int): Height and width of the map in pixels (it's a square).
        map_directory (str): Path to the directory where map files will be stored.
        logger (Any): Logger instance
    """

    def __init__(
        self,
        game: Game,
        dtm_provider: DTMProvider,
        dtm_provider_settings: DTMProviderSettings | None,
        coordinates: tuple[float, float],
        size: int,
        rotation: int,
        map_directory: str,
        logger: Any = None,
        custom_osm: str | None = None,
        dem_settings: DEMSettings = DEMSettings(),
        background_settings: BackgroundSettings = BackgroundSettings(),
        grle_settings: GRLESettings = GRLESettings(),
        i3d_settings: I3DSettings = I3DSettings(),
        texture_settings: TextureSettings = TextureSettings(),
        spline_settings: SplineSettings = SplineSettings(),
        satellite_settings: SatelliteSettings = SatelliteSettings(),
        **kwargs,
    ):
        if not logger:
            logger = Logger(to_stdout=True, to_file=False)
        self.logger = logger
        self.size = size

        if rotation:
            rotation_multiplier = 1.5
        else:
            rotation_multiplier = 1

        self.rotation = rotation
        self.rotated_size = int(size * rotation_multiplier)

        self.game = game
        self.dtm_provider = dtm_provider
        self.dtm_provider_settings = dtm_provider_settings
        self.components: list[Component] = []
        self.coordinates = coordinates
        self.map_directory = map_directory

        try:
            main_settings = {
                "game": game.code,
                "latitude": coordinates[0],
                "longitude": coordinates[1],
                "country": self.get_country_by_coordinates(),
                "size": size,
                "rotation": rotation,
                "dtm_provider": dtm_provider.name(),
                "custom_osm": bool(custom_osm),
                "is_public": kwargs.get("is_public", False),
            }
            send_main_settings(main_settings)
        except Exception as e:
            self.logger.error("Error sending main settings: %s", e)

        log_entry = ""
        log_entry += f"Map instance created for Game: {game.code}. "
        log_entry += f"Coordinates: {coordinates}. Size: {size}. Rotation: {rotation}. "
        log_entry += f"DTM provider is {dtm_provider.name()}. "

        self.custom_osm = custom_osm
        log_entry += f"Custom OSM file: {custom_osm}. "

        # Make a copy of a custom osm file to the map directory, so it will be
        # included in the output archive.
        if custom_osm:
            copy_path = os.path.join(self.map_directory, "custom_osm.osm")
            shutil.copyfile(custom_osm, copy_path)
            self.logger.debug("Custom OSM file copied to %s", copy_path)

        self.dem_settings = dem_settings
        log_entry += f"DEM settings: {dem_settings}. "
        if self.dem_settings.water_depth > 0:
            # Make sure that the plateau value is >= water_depth
            self.dem_settings.plateau = max(
                self.dem_settings.plateau, self.dem_settings.water_depth
            )

        self.background_settings = background_settings
        log_entry += f"Background settings: {background_settings}. "
        self.grle_settings = grle_settings
        log_entry += f"GRLE settings: {grle_settings}. "
        self.i3d_settings = i3d_settings
        log_entry += f"I3D settings: {i3d_settings}. "
        self.texture_settings = texture_settings
        log_entry += f"Texture settings: {texture_settings}. "
        self.spline_settings = spline_settings
        log_entry += f"Spline settings: {spline_settings}. "
        self.satellite_settings = satellite_settings

        self.logger.info(log_entry)
        os.makedirs(self.map_directory, exist_ok=True)
        self.logger.debug("Map directory created: %s", self.map_directory)

        settings = [
            dem_settings,
            background_settings,
            grle_settings,
            i3d_settings,
            texture_settings,
            spline_settings,
            satellite_settings,
        ]

        settings_json = {}

        for setting in settings:
            settings_json[setting.__class__.__name__] = setting.model_dump()

        try:
            send_advanced_settings(settings_json)
        except Exception as e:
            self.logger.error("Error sending advanced settings: %s", e)

        save_path = os.path.join(self.map_directory, "generation_settings.json")

        with open(save_path, "w", encoding="utf-8") as file:
            json.dump(settings_json, file, indent=4)

        self.shared_settings = SharedSettings()

        self.texture_custom_schema = kwargs.get("texture_custom_schema", None)
        if self.texture_custom_schema:
            save_path = os.path.join(self.map_directory, "texture_custom_schema.json")
            with open(save_path, "w", encoding="utf-8") as file:
                json.dump(self.texture_custom_schema, file, indent=4)
            self.logger.debug("Texture custom schema saved to %s", save_path)

        self.tree_custom_schema = kwargs.get("tree_custom_schema", None)
        if self.tree_custom_schema:
            save_path = os.path.join(self.map_directory, "tree_custom_schema.json")
            with open(save_path, "w", encoding="utf-8") as file:
                json.dump(self.tree_custom_schema, file, indent=4)
            self.logger.debug("Tree custom schema saved to %s", save_path)

        self.custom_background_path = kwargs.get("custom_background_path", None)
        if self.custom_background_path:
            save_path = os.path.join(self.map_directory, "custom_background.png")
            shutil.copyfile(self.custom_background_path, save_path)

        try:
            shutil.unpack_archive(game.template_path, self.map_directory)
            self.logger.debug("Map template unpacked to %s", self.map_directory)
        except Exception as e:
            raise RuntimeError(f"Can not unpack map template due to error: {e}") from e

    @property
    def texture_schema(self) -> list[dict[str, Any]] | None:
        """Return texture schema (custom if provided, default otherwise).

        Returns:
            list[dict[str, Any]] | None: Texture schema.
        """
        if self.texture_custom_schema:
            return self.texture_custom_schema
        with open(self.game.texture_schema, "r", encoding="utf-8") as file:
            return json.load(file)

    def generate(self) -> Generator[str, None, None]:
        """Launch map generation using all components. Yield component names during the process.

        Yields:
            Generator[str, None, None]: Component names.
        """
        self.logger.debug(
            "Starting map generation. Game code: %s. Coordinates: %s, size: %s. Rotation: %s.",
            self.game.code,
            self.coordinates,
            self.size,
            self.rotation,
        )

        for game_component in self.game.components:
            component = game_component(
                self.game,
                self,
                self.coordinates,
                self.size,
                self.rotated_size,
                self.rotation,
                self.map_directory,
                self.logger,
                texture_custom_schema=self.texture_custom_schema,
                tree_custom_schema=self.tree_custom_schema,
            )
            self.components.append(component)

            yield component.__class__.__name__

            try:
                component.process()
            except Exception as e:
                self.logger.error(
                    "Error processing component %s: %s",
                    component.__class__.__name__,
                    e,
                )
                raise e

            try:
                component.commit_generation_info()
            except Exception as e:
                self.logger.error(
                    "Error committing generation info for component %s: %s",
                    component.__class__.__name__,
                    e,
                )
                raise e

        self.logger.debug(
            "Map generation completed. Game code: %s. Coordinates: %s, size: %s. Rotation: %s.",
            self.game.code,
            self.coordinates,
            self.size,
            self.rotation,
        )

    def get_component(self, component_name: str) -> Component | None:
        """Get component by name.

        Arguments:
            component_name (str): Name of the component.

        Returns:
            Component | None: Component instance or None if not found.
        """
        for component in self.components:
            if component.__class__.__name__ == component_name:
                return component
        return None

    def get_texture_component(self) -> Texture | None:
        """Get texture component.

        Returns:
            Texture | None: Texture instance or None if not found.
        """
        component = self.get_component("Texture")
        if not isinstance(component, Texture):
            return None
        return component

    def get_background_component(self) -> Background | None:
        """Get background component.

        Returns:
            Background | None: Background instance or None if not found.
        """
        component = self.get_component("Background")
        if not isinstance(component, Background):
            return None
        return component

    def get_texture_layer(self, by_usage: str | None = None) -> Layer | None:
        """Get texture layer by usage.

        Arguments:
            by_usage (str, optional): Texture usage.

        Returns:
            Layer | None: Texture layer instance or None if not found.
        """
        texture_component = self.get_texture_component()
        if not texture_component:
            return None
        if by_usage:
            return texture_component.get_layer_by_usage(by_usage)
        return None

    def get_texture_layers(
        self,
        by_usage: str | None = None,
    ) -> None | list[Layer]:
        """Get texture layers by usage.

        Arguments:
            by_usage (str, optional): Texture usage.

        Returns:
            None | list[Layer]: List of texture layers.
        """
        texture_component = self.get_texture_component()
        if not texture_component:
            return None
        if by_usage:
            return texture_component.get_layers_by_usage(by_usage)
        return None

    def previews(self) -> list[str]:
        """Get list of preview images.

        Returns:
            list[str]: List of preview images.
        """
        previews = []
        for component in self.components:
            try:
                previews.extend(component.previews())
            except Exception as e:
                self.logger.error(
                    "Error getting previews for component %s: %s",
                    component.__class__.__name__,
                    e,
                )
        return previews

    def pack(self, archive_path: str, remove_source: bool = True) -> str:
        """Pack map directory to zip archive.

        Arguments:
            archive_path (str): Path to the archive.
            remove_source (bool, optional): Remove source directory after packing.

        Returns:
            str: Path to the archive.
        """
        archive_path = shutil.make_archive(archive_path, "zip", self.map_directory)
        self.logger.debug("Map packed to %s.zip", archive_path)
        if remove_source:
            try:
                shutil.rmtree(self.map_directory)
                self.logger.debug("Map directory removed: %s", self.map_directory)
            except Exception as e:
                self.logger.debug("Error removing map directory %s: %s", self.map_directory, e)
        return archive_path

    def get_country_by_coordinates(self) -> str:
        """Get country name by coordinates.

        Returns:
            str: Country name.
        """
        try:
            geolocator = Nominatim(user_agent="maps4fs")
            location = geolocator.reverse(self.coordinates, language="en")
            if location and "country" in location.raw["address"]:
                return location.raw["address"]["country"]
        except Exception as e:
            self.logger.error("Error getting country name by coordinates: %s", e)
            return "Unknown"
        return "Unknown"



"""Module for sending settings to the statistics server."""

import os
from typing import Any

import requests

from maps4fs.logger import Logger

logger = Logger()

try:
    from dotenv import load_dotenv

    load_dotenv("local.env")
except Exception:
    pass

STATS_HOST = os.getenv("STATS_HOST")
if not STATS_HOST:
    logger.debug("STATS_HOST not set in environment")
API_TOKEN = os.getenv("API_TOKEN")
if not API_TOKEN:
    logger.debug("API_TOKEN not set in environment")


def post(endpoint: str, data: dict[str, Any]) -> dict[str, Any] | None:
    """Make a POST request to the statistics server.

    Arguments:
        endpoint (str): The endpoint to send the request to.
        data (dict[str, Any]): The data to send.

    Returns:
        dict[str, Any]: The response from the server.
    """
    if not STATS_HOST or not API_TOKEN:
        logger.info("STATS_HOST or API_TOKEN not set in environment, can't send settings.")
        return None

    headers = {"Authorization": f"Bearer {API_TOKEN}", "Content-Type": "application/json"}
    response = requests.post(endpoint, headers=headers, json=data, timeout=10)
    if response.status_code != 200:
        logger.error("Failed to send settings: %s", response.text)
        return None
    logger.info("Settings sent successfully")
    return response.json()


def send_main_settings(data: dict[str, Any]) -> None:
    """Send main settings to the statistics server.

    Arguments:
        data (dict[str, Any]): The main settings to send.
    """
    endpoint = f"{STATS_HOST}/receive_main_settings"
    post(endpoint, data)


def send_advanced_settings(data: dict[str, Any]) -> None:
    """Send advanced settings to the statistics server.

    Arguments:
        data (dict[str, Any]): The advanced settings to send.
    """
    endpoint = f"{STATS_HOST}/receive_advanced_settings"
    post(endpoint, data)


def get_main_settings(fields: list[str], limit: int | None = None) -> list[dict[str, Any]] | None:
    """Get main settings from the statistics server.

    Arguments:
        fields (list[str]): The fields to get.
        limit (int | None): The maximum number of settings to get.

    Returns:
        list[dict[str, Any]]: The settings from the server.
    """
    endpoint = f"{STATS_HOST}/get_main_settings"
    data = {"fields": fields, "limit": limit}
    result = post(endpoint, data)
    if not result:
        return None
    return result.get("settings")



"""This module contains provider of USGS data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class USGSWCSProvider(WCSProvider, DTMProvider):
    """Provider of USGS data."""

    _code = "usgs_wcs"
    _name = "USGS"
    _region = "USA"
    _icon = "🇺🇸"
    _resolution = "1-90"
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _instructions = None
    _is_base = False
    _extents = [(50.0, 17.0, -64.0, -162.0)]

    _url = "https://elevation.nationalmap.gov/arcgis/services/3DEPElevation/ImageServer/WCSServer"
    _wcs_version = "1.0.0"
    _source_crs = "EPSG:3857"
    _tile_size = 1000
    _is_multipart = False

    def get_wcs_parameters(self, tile):
        return {
            "identifier": "DEP3Elevation",
            "bbox": (tile[1], tile[0], tile[3], tile[2]),
            "crs": "EPSG:3857",
            "width": 1000,
            "height": 1000,
            "format": "GeoTIFF",
        }



"""This module contains provider of Antarctic data."""

import os

import requests

from maps4fs.generator.dtm.dtm import DTMProvider


class REMAProvider(DTMProvider):
    """Provider of Antarctic data."""

    _code = "rema"
    _name = "REMA Antarctica"
    _region = "Global"
    _icon = "🌍"
    _resolution = 2
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True

    _extents = [(-53.5443873459092, -53.5443873459092, 179.99698443265999, -180)]

    _instructions = (
        "This provider source includes 2 meter DEM data for the entire Antarctic region below 53 "
        "degrees South. The tiles are very big, around 1 GB each, so downloading and processing "
        "them can take a long time."
    )

    _url = "https://stac.pgc.umn.edu/api/v1/collections/rema-mosaics-v2.0-2m/items"

    def download_tiles(self):
        download_urls = self.get_download_urls()
        all_tif_files = self.download_tif_files(download_urls, self.shared_tiff_path)
        return all_tif_files

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shared_tiff_path = os.path.join(self._tile_directory, "shared")
        os.makedirs(self.shared_tiff_path, exist_ok=True)

    def get_download_urls(self) -> list[str]:
        """Get download URLs of the GeoTIFF files from the OGC API.

        Returns:
            list: List of download URLs.
        """
        urls = []

        try:
            # Make the GET request
            north, south, east, west = self.get_bbox()
            response = requests.get(  # pylint: disable=W3101
                self.url,  # type: ignore
                params={
                    "bbox": f"{west},{south},{east},{north}",
                    "limit": "100",
                },
                timeout=60,
            )
            self.logger.debug("Getting file locations from REMA OGC API...")

            # Check if the request was successful (HTTP status code 200)
            if response.status_code == 200:
                # Parse the JSON response
                json_data = response.json()
                items = json_data["features"]
                for item in items:
                    urls.append(item["assets"]["dem"]["href"])
            else:
                self.logger.error("Failed to get data. HTTP Status Code: %s", response.status_code)
        except requests.exceptions.RequestException as e:
            self.logger.error("Failed to get data. Error: %s", e)
        self.logger.debug("Received %s urls", len(urls))
        return urls



"""This module contains provider of Mecklenburg-Vorpommern data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider, DTMProviderSettings


class MecklenburgVorpommernProviderSettings(DTMProviderSettings):
    """Settings for the Mecklenburg-Vorpommern provider."""

    dataset: dict | str = {
        "mv_dgm": "Mecklenburg-Vorpommern DGM1",
        "mv_dgm5": "Mecklenburg-Vorpommern DGM5",
        "mv_dgm25": "Mecklenburg-Vorpommern DGM25",
    }


class MecklenburgVorpommernProvider(WCSProvider, DTMProvider):
    """Provider of Mecklenburg-Vorpommern data."""

    _code = "mv"
    _name = "Mecklenburg-Vorpommern"
    _region = "DE"
    _icon = "🇩🇪"
    _resolution = "1-25"
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _instructions = None
    _is_base = False
    _settings = MecklenburgVorpommernProviderSettings
    _extents = [(54.8, 53, 14.5, 10.5)]

    _url = "https://www.geodaten-mv.de/dienste/dgm_wcs"
    _wcs_version = "2.0.1"
    _source_crs = "EPSG:25833"
    _tile_size = 1000

    def get_wcs_parameters(self, tile):
        return {
            "identifier": [self.user_settings.dataset],
            "subsets": [("x", str(tile[1]), str(tile[3])), ("y", str(tile[0]), str(tile[2]))],
            "format": "image/tiff",
        }



"""This module contains provider of Switzerland data."""

import json
import os

import requests

from maps4fs.generator.dtm import utils
from maps4fs.generator.dtm.dtm import DTMProvider, DTMProviderSettings


class SwitzerlandProviderSettings(DTMProviderSettings):
    """Settings for the Switzerland provider."""

    resolution: tuple | str = ("0.5", "2.0")


class SwitzerlandProvider(DTMProvider):
    """Provider of Switzerland."""

    _code = "switzerland"
    _name = "Switzerland"
    _region = "CH"
    _icon = "🇨🇭"
    _resolution = "0.5-2"
    _settings = SwitzerlandProviderSettings
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True

    _extents = [(47.8308275417, 45.7769477403, 10.4427014502, 6.02260949059)]

    _url = (
        "https://ogd.swisstopo.admin.ch/services/swiseld/"
        "services/assets/ch.swisstopo.swissalti3d/search"
    )

    def download_tiles(self):
        download_urls = self.get_download_urls()
        all_tif_files = self.download_tif_files(download_urls, self.shared_tiff_path)
        return all_tif_files

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shared_tiff_path = os.path.join(self._tile_directory, "shared")
        os.makedirs(self.shared_tiff_path, exist_ok=True)

    def get_download_urls(self) -> list[str]:
        """Get download URLs of the GeoTIFF files from the USGS API.

        Returns:
            list: List of download URLs.
        """
        urls = []
        try:
            bbox = self.get_bbox()
            north, south, east, west = utils.transform_bbox(bbox, "EPSG:2056")

            params = {
                "format": "image/tiff; application=geotiff; profile=cloud-optimized",
                "resolution": self.user_settings.resolution,  # type: ignore
                "srid": 2056,
                "state": "current",
            }

            data = {
                "geometry": json.dumps(
                    {
                        "type": "Polygon",
                        "crs": {"type": "name", "properties": {"name": "EPSG:2056"}},
                        "coordinates": [
                            [
                                [north, east],
                                [south, east],
                                [south, west],
                                [north, west],
                                [north, east],
                            ]
                        ],
                    }
                )
            }

            response = requests.post(  # pylint: disable=W3101
                self.url,  # type: ignore
                params=params,
                data=data,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                timeout=60,
            )
            self.logger.debug("Getting file locations...")

            # Check if the request was successful (HTTP status code 200)
            if response.status_code == 200:
                # Parse the JSON response
                json_data = response.json()
                items = json_data["items"]
                for item in items:
                    urls.append(item["ass_asset_href"])
            else:
                self.logger.error("Failed to get data. HTTP Status Code: %s", response.status_code)
        except requests.exceptions.RequestException as e:
            self.logger.error("Failed to get data. Error: %s", e)
        self.logger.debug("Received %s urls", len(urls))
        return urls



"""Base class for all components that primarily used to work with images."""

import os

import cv2
import numpy as np

from maps4fs.generator.component.base.component import Component
from maps4fs.generator.settings import Parameters


class ImageComponent(Component):
    """Base class for all components that primarily used to work with images."""

    @staticmethod
    def polygon_points_to_np(
        polygon_points: list[tuple[int, int]], divide: int | None = None
    ) -> np.ndarray:
        """Converts the polygon points to a NumPy array.

        Arguments:
            polygon_points (list[tuple[int, int]]): The polygon points.
            divide (int, optional): The number to divide the points by. Defaults to None.

        Returns:
            np.array: The NumPy array of the polygon points.
        """
        array = np.array(polygon_points, dtype=np.int32).reshape((-1, 1, 2))
        if divide:
            return array // divide
        return array

    @staticmethod
    def cut_out_np(
        image: np.ndarray, half_size: int, set_zeros: bool = False, return_cutout: bool = False
    ) -> np.ndarray:
        """Cuts out a square from the center of the image.

        Arguments:
            image (np.ndarray): The image.
            half_size (int): The half size of the square.
            set_zeros (bool, optional): Whether to set the cutout to zeros. Defaults to False.
            return_cutout (bool, optional): Whether to return the cutout. Defaults to False.

        Returns:
            np.ndarray: The image with the cutout or the cutout itself.
        """
        center = (image.shape[0] // 2, image.shape[1] // 2)
        x1 = center[0] - half_size
        x2 = center[0] + half_size
        y1 = center[1] - half_size
        y2 = center[1] + half_size

        if return_cutout:
            return image[x1:x2, y1:y2]

        if set_zeros:
            image[x1:x2, y1:y2] = 0

        return image

    @staticmethod
    def subtract_by_mask(
        image: np.ndarray,
        image_mask: np.ndarray,
        subtract_by: int,
        mask_by: int = 255,
        erode_kernel: int | None = 3,
        erode_iter: int | None = 1,
    ) -> np.ndarray:
        """Subtracts a value from the image where the mask is equal to the mask by value.

        Arguments:
            image (np.ndarray): The image.
            image_mask (np.ndarray): The mask.
            subtract_by (int): The value to subtract by.
            mask_by (int, optional): The value to mask by. Defaults to 255.
            erode_kernel (int, optional): The kernel size for the erosion. Defaults to 3.
            erode_iter (int, optional): The number of iterations for the erosion. Defaults to 1.

        Returns:
            np.ndarray: The image with the subtracted value.
        """
        mask = image_mask == mask_by
        if erode_kernel and erode_iter:
            mask = cv2.erode(
                mask.astype(np.uint8),
                np.ones((erode_kernel, erode_kernel), np.uint8),
                iterations=erode_iter,
            ).astype(bool)

        image[mask] = image[mask] - subtract_by
        return image

    @staticmethod
    def resize_to_preview(image_path: str, save_path: str) -> None:
        """Resizes an image to the preview size.

        Arguments:
            image_path (str): The path to the image.
            save_path (str): The path to save the resized image.
        """
        if not os.path.isfile(image_path):
            return

        image = cv2.imread(image_path)
        if image is None:
            return

        if (
            image.shape[0] > Parameters.PREVIEW_MAXIMUM_SIZE
            or image.shape[1] > Parameters.PREVIEW_MAXIMUM_SIZE
        ):
            image = cv2.resize(
                image, (Parameters.PREVIEW_MAXIMUM_SIZE, Parameters.PREVIEW_MAXIMUM_SIZE)
            )

        cv2.imwrite(save_path, image)

    @staticmethod
    def transfer_border(src_image: np.ndarray, dst_image: np.ndarray | None, border: int) -> None:
        """Transfers the border of the source image to the destination image.

        Arguments:
            src_image (np.ndarray): The source image.
            dst_image (np.ndarray, optional): The destination image.
            border (int): The border size.
        """
        borders = [
            (slice(None, border), slice(None)),
            (slice(None), slice(-border, None)),
            (slice(-border, None), slice(None)),
            (slice(None), slice(None, border)),
        ]

        for row_slice, col_slice in borders:
            border_slice = (row_slice, col_slice)
            if dst_image is not None:
                dst_image[border_slice][src_image[border_slice] != 0] = 255
            src_image[border_slice] = 0



"""Module with Texture class for generating textures for the map using OSM data."""

from __future__ import annotations

import json
import os
import shutil
import warnings
from collections import defaultdict
from typing import Any, Callable, Generator, Optional

import cv2
import numpy as np
import osmnx as ox
import pandas as pd
from osmnx import settings as ox_settings
from shapely import LineString, Point, Polygon
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from maps4fs.generator.component.base.component_image import ImageComponent
from maps4fs.generator.component.layer import Layer

PREVIEW_MAXIMUM_SIZE = 2048


class Texture(ImageComponent):
    """Class which generates textures for the map using OSM data.

    Attributes:
        weights_dir (str): Path to the directory with weights.
        name (str): Name of the texture.
        tags (dict[str, str | list[str] | bool]): Dictionary of tags to search for.
        width (int | None): Width of the polygon in meters (only for LineString).
        color (tuple[int, int, int]): Color of the layer in BGR format.
    """

    def preprocess(self) -> None:
        """Preprocesses the data before the generation."""
        self.read_layers(self.get_schema())

        self._weights_dir = self.game.weights_dir_path(self.map_directory)
        self.procedural_dir = os.path.join(self._weights_dir, "masks")
        os.makedirs(self.procedural_dir, exist_ok=True)

        self.info_save_path = os.path.join(self.map_directory, "generation_info.json")
        self.info_layer_path = os.path.join(self.info_layers_directory, "textures.json")

    def read_layers(self, layers_schema: list[dict[str, Any]]) -> None:
        """Reads layers from the schema.

        Arguments:
            layers_schema (list[dict[str, Any]]): Schema with layers for textures.
        """
        try:
            self.layers = [Layer.from_json(layer) for layer in layers_schema]
            self.logger.debug("Loaded %s layers.", len(self.layers))
        except Exception as e:
            raise ValueError(f"Error loading texture layers: {e}") from e

    def get_schema(self) -> list[dict[str, Any]]:
        """Returns schema with layers for textures.

        Raises:
            FileNotFoundError: If the schema file is not found.
            ValueError: If there is an error loading the schema.
            ValueError: If the schema is not a list of dictionaries.

        Returns:
            dict[str, Any]: Schema with layers for textures.
        """
        custom_schema = self.kwargs.get("texture_custom_schema")
        if custom_schema:
            layers_schema = custom_schema
            self.logger.debug("Custom schema loaded with %s layers.", len(layers_schema))
        else:
            if not os.path.isfile(self.game.texture_schema):
                raise FileNotFoundError(
                    f"Texture layers schema not found: {self.game.texture_schema}"
                )

            try:
                with open(self.game.texture_schema, "r", encoding="utf-8") as f:
                    layers_schema = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError(f"Error loading texture layers schema: {e}") from e

        if not isinstance(layers_schema, list):
            raise ValueError("Texture layers schema must be a list of dictionaries.")

        return layers_schema

    def get_base_layer(self) -> Layer | None:
        """Returns base layer.

        Returns:
            Layer | None: Base layer.
        """
        for layer in self.layers:
            if layer.priority == 0:
                return layer
        return None

    def get_background_layers(self) -> list[Layer]:
        """Returns list of background layers.

        Returns:
            list[Layer]: List of background layers.
        """
        return [layer for layer in self.layers if layer.background]

    def get_layer_by_usage(self, usage: str) -> Layer | None:
        """Returns layer by usage.

        Arguments:
            usage (str): Usage of the layer.

        Returns:
            Layer | None: Layer.
        """
        for layer in self.layers:
            if layer.usage == usage:
                return layer
        return None

    def get_layers_by_usage(self, usage: str) -> list[Layer]:
        """Returns layer by usage.

        Arguments:
            usage (str): Usage of the layer.

        Returns:
            list[Layer]: List of layers.
        """
        return [layer for layer in self.layers if layer.usage == usage]

    def process(self) -> None:
        """Processes the data to generate textures."""
        self._prepare_weights()
        self._read_parameters()
        self.draw()
        self.rotate_textures()
        self.add_borders()
        if self.map.texture_settings.dissolve and self.game.dissolve:
            self.dissolve()
        self.copy_procedural()

    def add_borders(self) -> None:
        """Iterates over all the layers and picks the one which have the border propety defined.
        Borders are distance from the edge of the map on each side (top, right, bottom, left).
        On the layers those pixels will be removed (value set to 0). If the base layer exist in
        the schema, those pixel values (not 0) will be added as 255 to the base layer."""
        base_layer = self.get_base_layer()
        base_layer_image = None
        if base_layer:
            base_layer_image = cv2.imread(base_layer.path(self._weights_dir), cv2.IMREAD_UNCHANGED)

        layers_with_borders = [layer for layer in self.layers if layer.border is not None]

        for layer in layers_with_borders:
            # Read the image.
            # Read pixels on borders with specified width (border property).
            # Where the pixel value is 255 - set it to 255 in base layer image.
            # And set it to 0 in the current layer image.
            layer_image = cv2.imread(layer.path(self._weights_dir), cv2.IMREAD_UNCHANGED)
            border = layer.border
            if not border:
                continue

            self.transfer_border(layer_image, base_layer_image, border)

            cv2.imwrite(layer.path(self._weights_dir), layer_image)
            self.logger.debug("Borders added to layer %s.", layer.name)

        if base_layer_image is not None:
            cv2.imwrite(base_layer.path(self._weights_dir), base_layer_image)  # type: ignore

    def copy_procedural(self) -> None:
        """Copies some of the textures to use them as mask for procedural generation.
        Creates an empty blockmask if it does not exist."""
        blockmask_path = os.path.join(self.procedural_dir, "BLOCKMASK.png")
        if not os.path.isfile(blockmask_path):
            self.logger.debug("BLOCKMASK.png not found, creating an empty file.")
            img = np.zeros((self.map_size, self.map_size), dtype=np.uint8)
            cv2.imwrite(blockmask_path, img)

        pg_layers_by_type = defaultdict(list)
        for layer in self.layers:
            if layer.procedural:
                # Get path to the original file.
                texture_path = layer.get_preview_or_path(self._weights_dir)
                for procedural_layer_name in layer.procedural:
                    pg_layers_by_type[procedural_layer_name].append(texture_path)

        if not pg_layers_by_type:
            self.logger.debug("No procedural layers found.")
            return

        for procedural_layer_name, texture_paths in pg_layers_by_type.items():
            procedural_save_path = os.path.join(self.procedural_dir, f"{procedural_layer_name}.png")
            if len(texture_paths) > 1:
                # If there are more than one texture, merge them.
                merged_texture = np.zeros((self.map_size, self.map_size), dtype=np.uint8)
                for texture_path in texture_paths:
                    texture = cv2.imread(texture_path, cv2.IMREAD_UNCHANGED)
                    merged_texture[texture == 255] = 255
                cv2.imwrite(procedural_save_path, merged_texture)
                self.logger.debug(
                    "Procedural file %s merged from %s textures.",
                    procedural_save_path,
                    len(texture_paths),
                )
            elif len(texture_paths) == 1:
                # Otherwise, copy the texture.
                shutil.copyfile(texture_paths[0], procedural_save_path)
                self.logger.debug(
                    "Procedural file %s copied from %s.", procedural_save_path, texture_paths[0]
                )

    def rotate_textures(self) -> None:
        """Rotates textures of the layers which have tags."""
        if self.rotation:
            # Iterate over the layers which have tags and rotate them.
            for layer in tqdm(self.layers, desc="Rotating textures", unit="layer"):
                if layer.tags or layer.precise_tags:
                    self.logger.debug("Rotating layer %s.", layer.name)
                    layer_paths = layer.paths(self._weights_dir)
                    layer_paths += [layer.path_preview(self._weights_dir)]
                    for layer_path in layer_paths:
                        if os.path.isfile(layer_path):
                            self.rotate_image(
                                layer_path,
                                self.rotation,
                                output_height=self.map_size,
                                output_width=self.map_size,
                            )
                else:
                    self.logger.debug(
                        "Skipping rotation of layer %s because it has no tags.", layer.name
                    )

    def _read_parameters(self) -> None:
        """Reads map parameters from OSM data, such as:
        - minimum and maximum coordinates
        """
        bbox = ox.utils_geo.bbox_from_point(self.coordinates, dist=self.map_rotated_size / 2)
        self.minimum_x, self.minimum_y, self.maximum_x, self.maximum_y = bbox

    def info_sequence(self) -> dict[str, Any]:
        """Returns the JSON representation of the generation info for textures."""
        useful_attributes = [
            "coordinates",
            "bbox",
            "map_size",
            "rotation",
            "minimum_x",
            "minimum_y",
            "maximum_x",
            "maximum_y",
        ]
        return {attr: getattr(self, attr, None) for attr in useful_attributes}

    def _prepare_weights(self):
        self.logger.debug("Starting preparing weights from %s layers.", len(self.layers))

        for layer in tqdm(self.layers, desc="Preparing weights", unit="layer"):
            self._generate_weights(layer)
        self.logger.debug("Prepared weights for %s layers.", len(self.layers))

    def _generate_weights(self, layer: Layer) -> None:
        """Generates weight files for textures. Each file is a numpy array of zeros and
            dtype uint8 (0-255).

        Arguments:
            layer (Layer): Layer with textures and tags.
        """
        if layer.tags is None and layer.precise_tags is None:
            size = (self.map_size, self.map_size)
        else:
            size = (self.map_rotated_size, self.map_rotated_size)
        postfix = "_weight.png" if not layer.exclude_weight else ".png"
        if layer.count == 0:
            filepaths = [os.path.join(self._weights_dir, layer.name + postfix)]
        else:
            filepaths = [
                os.path.join(self._weights_dir, layer.name + str(i).zfill(2) + postfix)
                for i in range(1, layer.count + 1)
            ]

        for filepath in filepaths:
            img = np.zeros(size, dtype=np.uint8)
            cv2.imwrite(filepath, img)

    @property
    def layers(self) -> list[Layer]:
        """Returns list of layers with textures and tags from textures.json.

        Returns:
            list[Layer]: List of layers.
        """
        return self._layers

    @layers.setter
    def layers(self, layers: list[Layer]) -> None:
        """Sets list of layers with textures and tags.

        Arguments:
            layers (list[Layer]): List of layers.
        """
        self._layers = layers

    def layers_by_priority(self) -> list[Layer]:
        """Returns list of layers sorted by priority: None priority layers are first,
        then layers are sorted by priority (descending).

        Returns:
            list[Layer]: List of layers sorted by priority.
        """
        return sorted(
            self.layers,
            key=lambda _layer: (
                _layer.priority is not None,
                -_layer.priority if _layer.priority is not None else float("inf"),
            ),
        )

    def draw(self) -> None:
        """Iterates over layers and fills them with polygons from OSM data."""
        layers = self.layers_by_priority()
        layers = [
            layer for layer in layers if layer.tags is not None or layer.precise_tags is not None
        ]

        cumulative_image = None

        # Dictionary to store info layer data.
        # Key is a layer.info_layer, value is a list of polygon points as tuples (x, y).
        info_layer_data: dict[str, list[list[int]]] = defaultdict(list)

        for layer in tqdm(layers, desc="Drawing textures", unit="layer"):
            if self.map.texture_settings.skip_drains and layer.usage == "drain":
                self.logger.debug("Skipping layer %s because of the usage.", layer.name)
                continue
            if layer.priority == 0:
                self.logger.debug(
                    "Found base layer %s. Postponing that to be the last layer drawn.", layer.name
                )
                continue
            layer_path = layer.path(self._weights_dir)
            self.logger.debug("Drawing layer %s.", layer_path)
            layer_image = cv2.imread(layer_path, cv2.IMREAD_UNCHANGED)

            if cumulative_image is None:
                self.logger.debug("First layer, creating new cumulative image.")
                cumulative_image = layer_image

            mask = cv2.bitwise_not(cumulative_image)
            self._draw_layer(layer, info_layer_data, layer_image)
            self._add_roads(layer, info_layer_data)

            output_image = cv2.bitwise_and(layer_image, mask)
            cumulative_image = cv2.bitwise_or(cumulative_image, output_image)

            cv2.imwrite(layer_path, output_image)
            self.logger.debug("Texture %s saved.", layer_path)

        # Save info layer data.
        if os.path.isfile(self.info_layer_path):
            self.logger.debug(
                "File %s already exists, will update to avoid overwriting.", self.info_layer_path
            )
            with open(self.info_layer_path, "r", encoding="utf-8") as f:
                info_layer_data.update(json.load(f))

        with open(self.info_layer_path, "w", encoding="utf-8") as f:
            json.dump(info_layer_data, f, ensure_ascii=False, indent=4)
            self.logger.debug("Info layer data saved to %s.", self.info_layer_path)

        if cumulative_image is not None:
            self.draw_base_layer(cumulative_image)

    def _draw_layer(
        self, layer: Layer, info_layer_data: dict[str, list[list[int]]], layer_image: np.ndarray
    ) -> None:
        """Draws polygons from OSM data on the layer image and updates the info layer data.

        Arguments:
            layer (Layer): Layer with textures and tags.
            info_layer_data (dict[list[list[int]]]): Dictionary to store info layer data.
            layer_image (np.ndarray): Layer image.
        """
        tags = layer.tags
        if self.map.texture_settings.use_precise_tags:
            if layer.precise_tags:
                self.logger.debug(
                    "Using precise tags: %s for layer %s.", layer.precise_tags, layer.name
                )
                tags = layer.precise_tags

        if tags is None:
            return

        for polygon in self.objects_generator(tags, layer.width, layer.info_layer):
            if not len(polygon) > 2:
                self.logger.debug("Skipping polygon with less than 3 points.")
                continue
            if layer.info_layer:
                info_layer_data[layer.info_layer].append(
                    self.np_to_polygon_points(polygon)  # type: ignore
                )
            if not layer.invisible:
                try:
                    cv2.fillPoly(layer_image, [polygon], color=255)  # type: ignore
                except Exception as e:
                    self.logger.warning("Error drawing polygon: %s.", repr(e))
                    continue

    def _add_roads(self, layer: Layer, info_layer_data: dict[str, list[list[int]]]) -> None:
        """Adds roads to the info layer data.

        Arguments:
            layer (Layer): Layer with textures and tags.
            info_layer_data (dict[list[list[int]]]): Dictionary to store info layer data.
        """
        if layer.info_layer == "roads":
            for linestring in self.objects_generator(
                layer.tags, layer.width, layer.info_layer, yield_linestrings=True
            ):
                linestring_entry = {
                    "points": linestring,
                    "tags": str(layer.tags),
                }
                info_layer_data[f"{layer.info_layer}_polylines"].append(linestring_entry)  # type: ignore

    def dissolve(self) -> None:
        """Dissolves textures of the layers with tags into sublayers for them to look more
        natural in the game.
        Iterates over all layers with tags and reads the first texture, checks if the file
        contains any non-zero values (255), splits those non-values between different weight
        files of the corresponding layer and saves the changes to the files.
        """
        for layer in tqdm(self.layers, desc="Dissolving textures", unit="layer"):
            if not layer.tags:
                self.logger.debug("Layer %s has no tags, there's nothing to dissolve.", layer.name)
                continue
            layer_path = layer.path(self._weights_dir)
            layer_paths = layer.paths(self._weights_dir)

            if len(layer_paths) < 2:
                self.logger.debug("Layer %s has only one texture, skipping.", layer.name)
                continue

            self.logger.debug("Dissolving layer from %s to %s.", layer_path, layer_paths)

            # Check if the image contains any non-zero values, otherwise continue.
            layer_image = cv2.imread(layer_path, cv2.IMREAD_UNCHANGED)

            if not np.any(layer_image):
                self.logger.debug(
                    "Layer %s does not contain any non-zero values, skipping.", layer.name
                )
                continue

            # Save the original image to use it for preview later, without combining the sublayers.
            cv2.imwrite(layer.path_preview(self._weights_dir), layer_image.copy())

            # Get the coordinates of non-zero values.
            non_zero_coords = np.column_stack(np.where(layer_image > 0))

            # Prepare sublayers.
            sublayers = [np.zeros_like(layer_image) for _ in range(layer.count)]

            # Randomly assign non-zero values to sublayers.
            for coord in non_zero_coords:
                sublayers[np.random.randint(0, layer.count)][coord[0], coord[1]] = 255

            # Save the sublayers.
            for sublayer, sublayer_path in zip(sublayers, layer_paths):
                cv2.imwrite(sublayer_path, sublayer)
                self.logger.debug("Sublayer %s saved.", sublayer_path)

            self.logger.debug("Dissolved layer %s.", layer.name)

    def draw_base_layer(self, cumulative_image: np.ndarray) -> None:
        """Draws base layer and saves it into the png file.
        Base layer is the last layer to be drawn, it fills the remaining area of the map.

        Arguments:
            cumulative_image (np.ndarray): Cumulative image with all layers.
        """
        base_layer = self.get_base_layer()
        if base_layer is not None:
            layer_path = base_layer.path(self._weights_dir)
            self.logger.debug("Drawing base layer %s.", layer_path)
            img = cv2.bitwise_not(cumulative_image)
            cv2.imwrite(layer_path, img)
            self.logger.debug("Base texture %s saved.", layer_path)

    def latlon_to_pixel(self, lat: float, lon: float) -> tuple[int, int]:
        """Converts latitude and longitude to pixel coordinates.

        Arguments:
            lat (float): Latitude.
            lon (float): Longitude.

        Returns:
            tuple[int, int]: Pixel coordinates.
        """
        x = int((lon - self.minimum_x) / (self.maximum_x - self.minimum_x) * self.map_rotated_size)
        y = int((lat - self.maximum_y) / (self.minimum_y - self.maximum_y) * self.map_rotated_size)
        return x, y

    def np_to_polygon_points(self, np_array: np.ndarray) -> list[tuple[int, int]]:
        """Converts numpy array of polygon points to list of tuples.

        Arguments:
            np_array (np.ndarray): Numpy array of polygon points.

        Returns:
            list[tuple[int, int]]: List of tuples.
        """
        return [(int(x), int(y)) for x, y in np_array.reshape(-1, 2)]

    def _to_np(self, geometry: Polygon, *args) -> np.ndarray:
        """Converts Polygon geometry to numpy array of polygon points.

        Arguments:
            geometry (Polygon): Polygon geometry.
            *Arguments: Additional arguments:
                - width (int | None): Width of the polygon in meters.

        Returns:
            np.ndarray: Numpy array of polygon points.
        """
        coords = list(geometry.exterior.coords)
        pts = np.array(coords, np.int32)
        pts = pts.reshape((-1, 1, 2))
        return pts

    def _to_polygon(self, obj: pd.core.series.Series, width: int | None) -> Polygon:
        """Converts OSM object to numpy array of polygon points and converts coordinates to pixels.

        Arguments:
            obj (pd.core.series.Series): OSM object.
            width (int | None): Width of the polygon in meters.

        Returns:
            Polygon: Polygon geometry with pixel coordinates.
        """
        geometry = obj["geometry"]
        geometry_type = geometry.geom_type
        converter = self._converters(geometry_type)
        if not converter:
            self.logger.debug("Geometry type %s not supported.", geometry_type)
            return None
        return converter(geometry, width)

    def polygon_to_pixel_coordinates(self, polygon: Polygon) -> Polygon:
        """Converts polygon coordinates from lat lon to pixel coordinates.

        Arguments:
            polygon (Polygon): Polygon geometry.

        Returns:
            Polygon: Polygon geometry.
        """
        coords_pixel = [
            self.latlon_to_pixel(lat, lon) for lon, lat in list(polygon.exterior.coords)
        ]
        return Polygon(coords_pixel)

    def linestring_to_pixel_coordinates(self, linestring: LineString) -> LineString:
        """Converts LineString coordinates from lat lon to pixel coordinates.

        Arguments:
            linestring (LineString): LineString geometry.

        Returns:
            LineString: LineString geometry.
        """
        coords_pixel = [self.latlon_to_pixel(lat, lon) for lon, lat in list(linestring.coords)]
        return LineString(coords_pixel)

    def point_to_pixel_coordinates(self, point: Point) -> Point:
        """Converts Point coordinates from lat lon to pixel coordinates.

        Arguments:
            point (Point): Point geometry.

        Returns:
            Point: Point geometry.
        """
        x, y = self.latlon_to_pixel(point.y, point.x)
        return Point(x, y)

    def _to_pixel(self, geometry: Polygon, *args, **kwargs) -> Polygon:
        """Returns the same geometry with pixel coordinates.

        Arguments:
            geometry (Polygon): Polygon geometry.

        Returns:
            Polygon: Polygon geometry with pixel coordinates.
        """
        return self.polygon_to_pixel_coordinates(geometry)

    def _sequence_to_pixel(
        self,
        geometry: LineString | Point,
        width: int | None,
    ) -> Polygon:
        """Converts LineString or Point geometry to numpy array of polygon points.

        Arguments:
            geometry (LineString | Point): LineString or Point geometry.
            width (int | None): Width of the polygon in meters.

        Raises:
            ValueError: If the geometry type is not supported

        Returns:
            Polygon: Polygon geometry.
        """
        if isinstance(geometry, LineString):
            geometry = self.linestring_to_pixel_coordinates(geometry)
        elif isinstance(geometry, Point):
            geometry = self.point_to_pixel_coordinates(geometry)
        else:
            raise ValueError(f"Geometry type {type(geometry)} not supported.")

        return geometry.buffer(width if width else 0)

    def _converters(
        self, geom_type: str
    ) -> Optional[Callable[[BaseGeometry, Optional[int]], np.ndarray]]:
        """Returns a converter function for a given geometry type.

        Arguments:
            geom_type (str): Geometry type.

        Returns:
            Callable[[shapely.geometry, int | None], np.ndarray]: Converter function.
        """
        converters = {
            "Polygon": self._to_pixel,
            "LineString": self._sequence_to_pixel,
            "Point": self._sequence_to_pixel,
        }
        return converters.get(geom_type)  # type: ignore

    def objects_generator(
        self,
        tags: dict[str, str | list[str] | bool] | None,
        width: int | None,
        info_layer: str | None = None,
        yield_linestrings: bool = False,
    ) -> Generator[np.ndarray, None, None] | Generator[list[tuple[int, int]], None, None]:
        """Generator which yields numpy arrays of polygons from OSM data.

        Arguments:
            tags (dict[str, str | list[str]]): Dictionary of tags to search for.
            width (int | None): Width of the polygon in meters (only for LineString).
            info_layer (str | None): Name of the corresponding info layer.
            yield_linestrings (bool): Flag to determine if the LineStrings should be yielded.

        Yields:
            Generator[np.ndarray, None, None] | Generator[list[tuple[int, int]], None, None]:
                Numpy array of polygon points or list of point coordinates.
        """
        if tags is None:
            return
        is_fieds = info_layer == "fields"

        ox_settings.use_cache = self.map.texture_settings.use_cache
        ox_settings.requests_timeout = 30

        try:
            if self.map.custom_osm is not None:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", FutureWarning)
                    objects = ox.features_from_xml(self.map.custom_osm, tags=tags)
            else:
                objects = ox.features_from_bbox(bbox=self.new_bbox, tags=tags)
        except Exception as e:
            self.logger.debug("Error fetching objects for tags: %s. Error: %s.", tags, e)
            return
        self.logger.debug("Fetched %s elements for tags: %s.", len(objects), tags)

        method = self.linestrings_generator if yield_linestrings else self.polygons_generator

        yield from method(objects, width, is_fieds)

    def linestrings_generator(
        self, objects: pd.core.frame.DataFrame, *args, **kwargs
    ) -> Generator[list[tuple[int, int]], None, None]:
        """Generator which yields lists of point coordinates which represent LineStrings from OSM.

        Arguments:
            objects (pd.core.frame.DataFrame): Dataframe with OSM objects.

        Yields:
            Generator[list[tuple[int, int]], None, None]: List of point coordinates.
        """
        for _, obj in objects.iterrows():
            geometry = obj["geometry"]
            if isinstance(geometry, LineString):
                points = [self.latlon_to_pixel(x, y) for y, x in geometry.coords]
                yield points

    def polygons_generator(
        self, objects: pd.core.frame.DataFrame, width: int | None, is_fieds: bool
    ) -> Generator[np.ndarray, None, None]:
        """Generator which yields numpy arrays of polygons from OSM data.

        Arguments:
            objects (pd.core.frame.DataFrame): Dataframe with OSM objects.
            width (int | None): Width of the polygon in meters (only for LineString).
            is_fieds (bool): Flag to determine if the fields should be padded.

        Yields:
            Generator[np.ndarray, None, None]: Numpy array of polygon points.
        """
        for _, obj in objects.iterrows():
            try:
                polygon = self._to_polygon(obj, width)
            except Exception as e:
                self.logger.warning("Error converting object to polygon: %s.", e)
                continue
            if polygon is None:
                continue

            if is_fieds and self.map.texture_settings.fields_padding > 0:
                padded_polygon = polygon.buffer(-self.map.texture_settings.fields_padding)

                if not isinstance(padded_polygon, Polygon) or not list(
                    padded_polygon.exterior.coords
                ):
                    self.logger.debug("The padding value is too high, field will not padded.")
                else:
                    polygon = padded_polygon

            polygon_np = self._to_np(polygon)
            yield polygon_np

    def previews(self) -> list[str]:
        """Invokes methods to generate previews. Returns list of paths to previews.

        Returns:
            list[str]: List of paths to previews.
        """
        preview_paths = []
        preview_paths.append(self._osm_preview())
        return preview_paths

    def _osm_preview(self) -> str:
        """Merges layers into one image and saves it into the png file.

        Returns:
            str: Path to the preview.
        """
        scaling_factor = PREVIEW_MAXIMUM_SIZE / self.map_size

        preview_size = (
            int(self.map_size * scaling_factor),
            int(self.map_size * scaling_factor),
        )
        self.logger.debug(
            "Scaling factor: %s. Preview size: %s.",
            scaling_factor,
            preview_size,
        )

        active_layers = [
            layer
            for layer in self.layers
            if layer.tags is not None or layer.precise_tags is not None
        ]
        self.logger.debug("Following layers have tag textures: %s.", len(active_layers))

        images = [
            cv2.resize(
                cv2.imread(layer.get_preview_or_path(self._weights_dir), cv2.IMREAD_UNCHANGED),
                preview_size,
            )
            for layer in active_layers
        ]
        colors = [layer.color for layer in active_layers]
        color_images = []
        for img, color in zip(images, colors):
            color_img = np.zeros((img.shape[0], img.shape[1], 3), dtype=np.uint8)
            color_img[img > 0] = color
            color_images.append(color_img)
        merged = np.sum(color_images, axis=0, dtype=np.uint8)
        self.logger.debug(
            "Merged layers into one image. Shape: %s, dtype: %s.",
            merged.shape,
            merged.dtype,
        )
        preview_path = os.path.join(self.previews_directory, "textures_osm.png")

        cv2.imwrite(preview_path, merged)  # type: ignore
        self.logger.debug("Preview saved to %s.", preview_path)
        return preview_path



"""Base class for all components that primarily used to work with XML files."""

import os
from xml.etree import ElementTree as ET

from maps4fs.generator.component.base.component import Component


class XMLComponent(Component):
    """Base class for all components that primarily used to work with XML files."""

    xml_path: str | None = None

    def get_tree(self, xml_path: str | None = None) -> ET.ElementTree:
        """Parses the XML file and returns the tree.
        If the XML path is not provided, the class attribute is used.

        Arguments:
            xml_path (str, optional): The path to the XML file. Defaults to None.

        Raises:
            ValueError: If the XML path is not set as a class attribute nor provided as an
                argument.
            FileNotFoundError: If the XML file does not exist

        Returns:
            ET.ElementTree: The parsed XML tree.
        """
        xml_path = xml_path or self.xml_path
        if not xml_path:
            raise ValueError(
                "XML path was not set as a class attribute nor provided as an argument."
            )

        if not os.path.isfile(xml_path):
            raise FileNotFoundError(f"XML file {xml_path} does not exist.")

        return ET.parse(xml_path)

    def save_tree(self, tree: ET.ElementTree, xml_path: str | None = None) -> None:
        """Saves the XML tree to the file.
        If the XML path is not provided, the class attribute is used.

        Arguments:
            tree (ET.ElementTree): The XML tree to save.
            xml_path (str, optional): The path to the XML file. Defaults to None.

        Raises:
            ValueError: If the XML path is not set as a class attribute nor provided as an
                argument.
        """
        xml_path = xml_path or self.xml_path
        if not xml_path:
            raise ValueError(
                "XML path was not set as a class attribute nor provided as an argument."
            )

        tree.write(xml_path, encoding="utf-8", xml_declaration=True)

    def get_and_update_element(self, root: ET.Element, path: str, data: dict[str, str]) -> None:
        """Finds the element by the path and updates it with the provided data.

        Arguments:
            root (ET.Element): The root element of the XML tree.
            path (str): The path to the element.
            data (dict[str, str]): The data to update the element with.
        """
        element = root.find(path)
        if element is not None:
            self.update_element(element, data)

    def update_element(self, element: ET.Element, data: dict[str, str]) -> None:
        """Updates the element with the provided data.

        Arguments:
            element (ET.Element): The element to update.
            data (dict[str, str]): The data to update the element with.
        """
        for key, value in data.items():
            element.set(key, value)

    def create_element(self, element_name: str, data: dict[str, str]) -> ET.Element:
        """Creates an element with the provided data.

        Arguments:
            element_name (str): The name of the element.
            data (dict[str, str]): The data to set the element attributes to.

        Returns:
            ET.Element: The created element.
        """
        element = ET.Element(element_name)
        for key, value in data.items():
            element.set(key, value)
        return element

    def create_subelement(
        self, parent: ET.Element, element_name: str, data: dict[str, str]
    ) -> None:
        """Creates a subelement under the parent element with the provided data.

        Arguments:
            parent (ET.Element): The parent element.
            element_name (str): The name of the subelement.
            data (dict[str, str]): The data to set the subelement attributes to.
        """
        element = ET.SubElement(parent, element_name)
        self.update_element(element, data)



"""This module contains settings models for all components."""

from __future__ import annotations

import re
from typing import Any

from pydantic import BaseModel, ConfigDict


class Parameters:
    """Simple class to store string constants for parameters."""

    FIELD = "field"
    FIELDS = "fields"
    BUILDINGS = "buildings"
    TEXTURES = "textures"
    FOREST = "forest"
    ROADS_POLYLINES = "roads_polylines"
    FARMYARDS = "farmyards"

    PREVIEW_MAXIMUM_SIZE = 2048

    BACKGROUND_DISTANCE = 2048
    FULL = "FULL"
    PREVIEW = "PREVIEW"


class SharedSettings(BaseModel):
    """Represents the shared settings for all components."""

    mesh_z_scaling_factor: float | None = None
    height_scale_multiplier: float | None = None
    height_scale_value: float | None = None
    change_height_scale: bool = False

    model_config = ConfigDict(
        frozen=False,
    )


class SettingsModel(BaseModel):
    """Base class for settings models. It provides methods to convert settings to and from JSON."""

    model_config = ConfigDict(
        frozen=False,
    )

    @classmethod
    def all_settings_to_json(cls) -> dict[str, dict[str, Any]]:
        """Get all settings of the current class and its subclasses as a dictionary.

        Returns:
            dict[str, dict[str, Any]]: Dictionary with settings of the current class and its
                subclasses.
        """
        all_settings = {}
        for subclass in cls.__subclasses__():
            all_settings[subclass.__name__] = subclass().model_dump()

        return all_settings

    @classmethod
    def all_settings_from_json(
        cls, data: dict, flattening: bool = True
    ) -> dict[str, SettingsModel]:
        """Create settings instances from JSON data.

        Arguments:
            data (dict): JSON data.
            flattening (bool): if set to True will flattet iterables to use the first element
                of it.

        Returns:
            dict[str, Type[SettingsModel]]: Dictionary with settings instances.
        """
        settings = {}
        for subclass in cls.__subclasses__():
            subclass_data = data[subclass.__name__]
            if flattening:
                for key, value in subclass_data.items():
                    if isinstance(value, (list, tuple)):
                        subclass_data[key] = value[0]

            settings[cls.camel_to_snake(subclass.__name__)] = subclass(**subclass_data)

        return settings

    @staticmethod
    def camel_to_snake(camel_string: str) -> str:
        """Convert a camel case string to snake case.

        Arguments:
            camel_string (str): Camel case string.

        Returns:
            str: Snake case string.
        """
        splitted = re.split(r"(Settings)", camel_string)
        joined = "_".join(part.lower() for part in splitted if part)
        return joined

    @classmethod
    def all_settings(cls) -> list[SettingsModel]:
        """Get all settings of the current class and its subclasses.

        Returns:
            list[SettingsModel]: List with settings of the current class and its subclasses.
        """
        settings = []
        for subclass in cls.__subclasses__():
            settings.append(subclass())

        return settings


class DEMSettings(SettingsModel):
    """Represents the advanced settings for DEM component.

    Attributes:
        adjust_terrain_to_ground_level (bool): adjust terrain to ground level or not.
        multiplier (int): multiplier for the heightmap.
        blur_radius (int): radius of the blur filter.
        minimum_height_scale (int): minimum height scale for the i3d.
        plateau (int): plateau height.
        ceiling (int): ceiling height.
        water_depth (int): water depth.
    """

    adjust_terrain_to_ground_level: bool = True
    multiplier: int = 1
    minimum_height_scale: int = 255
    plateau: int = 0
    ceiling: int = 0
    water_depth: int = 0
    blur_radius: int = 3
    add_foundations: bool = False


class BackgroundSettings(SettingsModel):
    """Represents the advanced settings for background component.

    Attributes:
        generate_background (bool): generate obj files for the background terrain.
        generate_water (bool): generate obj files for the water.
        resize_factor (int): resize factor for the background terrain and water.
            It will be used as 1 / resize_factor of the original size.
    """

    generate_background: bool = False
    generate_water: bool = False
    resize_factor: int = 8
    remove_center: bool = True
    apply_decimation: bool = False
    decimation_percent: int = 25
    decimation_agression: int = 3


class GRLESettings(SettingsModel):
    """Represents the advanced settings for GRLE component.

    Attributes:
        farmland_margin (int): margin around the farmland.
        random_plants (bool): generate random plants on the map or use the default one.
        add_farmyards (bool): If True, regions of frarmyards will be added to the map
            without corresponding fields.
    """

    farmland_margin: int = 0
    add_farmyards: bool = False
    base_price: int = 60000
    price_scale: int = 100
    add_grass: bool = True
    base_grass: tuple | str = ("smallDenseMix", "meadow")
    random_plants: bool = True
    plants_island_minimum_size: int = 10
    plants_island_maximum_size: int = 200
    plants_island_vertex_count: int = 30
    plants_island_rounding_radius: int = 15
    plants_island_percent: int = 100
    fill_empty_farmlands: bool = False


class I3DSettings(SettingsModel):
    """Represents the advanced settings for I3D component.

    Attributes:
        forest_density (int): density of the forest (distance between trees).
    """

    add_trees: bool = True
    forest_density: int = 10
    trees_relative_shift: int = 20


class TextureSettings(SettingsModel):
    """Represents the advanced settings for texture component.

    Attributes:
        dissolve (bool): dissolve the texture into several images.
        fields_padding (int): padding around the fields.
        skip_drains (bool): skip drains generation.
    """

    dissolve: bool = False
    fields_padding: int = 0
    skip_drains: bool = False
    use_cache: bool = True
    use_precise_tags: bool = False


class SplineSettings(SettingsModel):
    """Represents the advanced settings for spline component.

    Attributes:
        spline_density (int): the number of extra points that will be added between each two
            existing points.
    """

    spline_density: int = 2
    add_reversed_splines: bool = False


class SatelliteSettings(SettingsModel):
    """Represents the advanced settings for satellite component.

    Attributes:
        download_images (bool): download satellite images.
        margin (int): margin around the map.
    """

    download_images: bool = False
    satellite_margin: int = 0
    zoom_level: int = 16



"""This module contains the Game class and its subclasses. Game class is used to define
different versions of the game for which the map is generated. Each game has its own map
template file and specific settings for map generation."""

from __future__ import annotations

import os

from maps4fs.generator.component.background import Background
from maps4fs.generator.component.config import Config
from maps4fs.generator.component.grle import GRLE
from maps4fs.generator.component.i3d import I3d
from maps4fs.generator.component.satellite import Satellite
from maps4fs.generator.component.texture import Texture

working_directory = os.getcwd()


class Game:
    """Class used to define different versions of the game for which the map is generated.

    Arguments:
        map_template_path (str, optional): Path to the map template file. Defaults to None.

    Attributes and Properties:
        code (str): The code of the game.
        components (list[Type[Component]]): List of components used for map generation.
        map_template_path (str): Path to the map template file.

    Public Methods:
        from_code(cls, code: str) -> Game: Returns the game instance based on the game code.
    """

    code: str | None = None
    dem_multipliyer: int = 2
    _additional_dem_name: str | None = None
    _map_template_path: str | None = None
    _texture_schema: str | None = None
    _grle_schema: str | None = None
    _tree_schema: str | None = None
    _i3d_processing: bool = True
    _plants_processing: bool = True
    _dissolve: bool = True

    # Order matters! Some components depend on others.
    components = [Texture, Background, GRLE, I3d, Config, Satellite]

    def __init__(self, map_template_path: str | None = None):
        if map_template_path:
            self._map_template_path = map_template_path

    def map_xml_path(self, map_directory: str) -> str:
        """Returns the path to the map.xml file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the map.xml file.
        """
        return os.path.join(map_directory, "maps", "map", "map.xml")

    @classmethod
    def from_code(cls, code: str, map_template_path: str | None = None) -> Game:
        """Returns the game instance based on the game code.

        Arguments:
            code (str): The code of the game.
            map_template_path (str, optional): Path to the map template file. Defaults to None.

        Returns:
            Game: The game instance.
        """
        for game in cls.__subclasses__():
            if game.code and game.code.lower() == code.lower():
                return game(map_template_path)
        raise ValueError(f"Game with code {code} not found.")

    @property
    def template_path(self) -> str:
        """Returns the path to the map template file.

        Raises:
            ValueError: If the map template path is not set.

        Returns:
            str: The path to the map template file."""
        if not self._map_template_path:
            raise ValueError("Map template path not set.")
        return self._map_template_path

    @property
    def texture_schema(self) -> str:
        """Returns the path to the texture layers schema file.

        Raises:
            ValueError: If the texture layers schema path is not set.

        Returns:
            str: The path to the texture layers schema file."""
        if not self._texture_schema:
            raise ValueError("Texture layers schema path not set.")
        return self._texture_schema

    @property
    def grle_schema(self) -> str:
        """Returns the path to the GRLE layers schema file.

        Raises:
            ValueError: If the GRLE layers schema path is not set.

        Returns:
            str: The path to the GRLE layers schema file."""
        if not self._grle_schema:
            raise ValueError("GRLE layers schema path not set.")
        return self._grle_schema

    @property
    def tree_schema(self) -> str:
        """Returns the path to the tree layers schema file.

        Raises:
            ValueError: If the tree layers schema path is not set.

        Returns:
            str: The path to the tree layers schema file."""
        if not self._tree_schema:
            raise ValueError("Tree layers schema path not set.")
        return self._tree_schema

    def dem_file_path(self, map_directory: str) -> str:
        """Returns the path to the DEM file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the DEM file.
        """
        raise NotImplementedError

    def weights_dir_path(self, map_directory: str) -> str:
        """Returns the path to the weights directory.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the weights directory."""
        raise NotImplementedError

    def get_density_map_fruits_path(self, map_directory: str) -> str:
        """Returns the path to the density map fruits file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the density map fruits file."""
        weights_dir = self.weights_dir_path(map_directory)
        return os.path.join(weights_dir, "densityMap_fruits.png")

    def get_farmlands_path(self, map_directory: str) -> str:
        """Returns the path to the farmlands file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the farmlands file."""
        weights_dir = self.weights_dir_path(map_directory)
        return os.path.join(weights_dir, "infoLayer_farmlands.png")

    def get_farmlands_xml_path(self, map_directory: str) -> str:
        """Returns the path to the farmlands xml file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the farmlands xml file."""
        raise NotImplementedError

    def i3d_file_path(self, map_directory: str) -> str:
        """Returns the path to the i3d file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the i3d file."""
        raise NotImplementedError

    @property
    def i3d_processing(self) -> bool:
        """Returns whether the i3d file should be processed.

        Returns:
            bool: True if the i3d file should be processed, False otherwise."""
        return self._i3d_processing

    @property
    def plants_processing(self) -> bool:
        """Returns whether the plants should be processed.

        Returns:
            bool: True if the plants should be processed, False otherwise."""
        return self._plants_processing

    @property
    def additional_dem_name(self) -> str | None:
        """Returns the name of the additional DEM file.

        Returns:
            str | None: The name of the additional DEM file."""
        return self._additional_dem_name

    def splines_file_path(self, map_directory: str) -> str:
        """Returns the path to the splines file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the splines file."""
        i3d_base_directory = os.path.dirname(self.i3d_file_path(map_directory))
        return os.path.join(i3d_base_directory, "splines.i3d")

    @property
    def dissolve(self) -> bool:
        """Returns whether the dissolve should be applied.

        Returns:
            bool: True if the dissolve should be applied, False otherwise."""
        return self._dissolve


class FS22(Game):
    """Class used to define the game version FS22."""

    code = "FS22"
    _map_template_path = os.path.join(working_directory, "data", "fs22-map-template.zip")
    _texture_schema = os.path.join(working_directory, "data", "fs22-texture-schema.json")
    _i3d_processing = False
    _plants_processing = False
    _dissolve = False

    def dem_file_path(self, map_directory: str) -> str:
        """Returns the path to the DEM file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the DEM file."""
        return os.path.join(map_directory, "maps", "map", "data", "map_dem.png")

    def weights_dir_path(self, map_directory: str) -> str:
        """Returns the path to the weights directory.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the weights directory."""
        return os.path.join(map_directory, "maps", "map", "data")

    def i3d_file_path(self, map_directory: str) -> str:
        """Returns the path to the i3d file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the i3d file."""
        return os.path.join(map_directory, "maps", "map", "map.i3d")


class FS25(Game):
    """Class used to define the game version FS25."""

    code = "FS25"
    dem_multipliyer: int = 2
    _additional_dem_name = "unprocessedHeightMap.png"
    _map_template_path = os.path.join(working_directory, "data", "fs25-map-template.zip")
    _texture_schema = os.path.join(working_directory, "data", "fs25-texture-schema.json")
    _grle_schema = os.path.join(working_directory, "data", "fs25-grle-schema.json")
    _tree_schema = os.path.join(working_directory, "data", "fs25-tree-schema.json")

    def dem_file_path(self, map_directory: str) -> str:
        """Returns the path to the DEM file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the DEM file."""
        return os.path.join(map_directory, "map", "data", "dem.png")

    def map_xml_path(self, map_directory: str) -> str:
        """Returns the path to the map.xml file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the map.xml file.
        """
        return os.path.join(map_directory, "map", "map.xml")

    def weights_dir_path(self, map_directory: str) -> str:
        """Returns the path to the weights directory.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the weights directory."""
        return os.path.join(map_directory, "map", "data")

    def i3d_file_path(self, map_directory: str) -> str:
        """Returns the path to the i3d file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the i3d file."""
        return os.path.join(map_directory, "map", "map.i3d")

    def get_farmlands_xml_path(self, map_directory: str) -> str:
        """Returns the path to the farmlands xml file.

        Arguments:
            map_directory (str): The path to the map directory.

        Returns:
            str: The path to the farmlands xml file."""
        return os.path.join(map_directory, "map", "config", "farmlands.xml")



"""This module contains provider of Czech data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class CzechProvider(WCSProvider, DTMProvider):
    """Provider of Czech data."""

    _code = "czech"
    _name = "Czech Republic"
    _region = "CZ"
    _icon = "🇨🇿"
    _resolution = 5
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _instructions = None
    _is_base = False
    _extents = [
        (51.0576876059846754, 48.4917065572081754, 18.9775933665038821, 12.0428143585602161)
    ]

    _url = "https://ags.cuzk.cz/arcgis2/services/INSPIRE_Nadmorska_vyska/ImageServer/WCSServer"  # pylint: disable=line-too-long
    _wcs_version = "1.0.0"
    _source_crs = "EPSG:4326"
    _tile_size = 0.05

    def get_wcs_parameters(self, tile):
        return {
            "identifier": "MD_LAS",
            "crs": "EPSG:4326",
            "bbox": tile,
            "width": 1000,
            "height": 1000,
            "format": "GeoTIFF",
        }



"""This module contains provider of Baden-Württemberg data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class BadenWurttembergProvider(WCSProvider, DTMProvider):
    """Provider of Baden-Württemberg data."""

    _code = "baden"
    _name = "Baden-Württemberg"
    _region = "DE"
    _icon = "🇩🇪"
    _resolution = "1"
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _instructions = None
    _is_base = False
    _extents = [(49.79645444804715, 47.52877040346605, 10.54203149250156, 7.444081717803481)]

    _url = "https://owsproxy.lgl-bw.de/owsproxy/wcs/WCS_INSP_BW_Hoehe_Coverage_DGM1"
    _wcs_version = "2.0.1"
    _source_crs = "EPSG:25832"
    _tile_size = 1000

    def get_wcs_parameters(self, tile):
        return {
            "identifier": ["EL.ElevationGridCoverage"],
            "subsets": [("E", str(tile[1]), str(tile[3])), ("N", str(tile[0]), str(tile[2]))],
            "format": "image/tiff",
        }



"""This module contains provider of Niedersachsen data."""

from maps4fs.generator.dtm.base.wms import WMSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class NiedersachsenProvider(WMSProvider, DTMProvider):
    """Provider of Niedersachsen data."""

    _code = "niedersachsen"
    _name = "Lower Saxony DGM1"
    _region = "DE"
    _icon = "🇩🇪󠁥󠁢󠁹󠁿"
    _resolution = 1
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _instructions = (
        "Warning: The Niedersachsen DGM1 data is provided as 8-bit Cloud Optimized GeoTIFF "
        "(whole meters only). You will need to use blur ('Blur Radius' under 'DEM Settings') "
        "to smooth the data."
    )
    _is_base = False
    _extents = [(54.148101, 51.153098, 11.754046, 6.505772)]

    _url = "https://opendata.lgln.niedersachsen.de/doorman/noauth/dgm_wms"
    _source_crs = "EPSG:25832"
    _tile_size = 2000
    _wms_version = "1.3.0"

    def get_wms_parameters(self, tile):
        return {
            "layers": ["ni_dgm1_grau"],
            "srs": "EPSG:25832",
            "bbox": (tile[1], tile[0], tile[3], tile[2]),
            "size": (2000, 2000),
            "format": "image/tiff",
            "transparent": False,
        }



"""This module contains provider of Finland data."""

from owslib.util import Authentication

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider, DTMProviderSettings


class FinlandProviderSettings(DTMProviderSettings):
    """Settings for the Finland provider."""

    api_key: str = ""


class FinlandProvider(WCSProvider, DTMProvider):
    """Provider of Finland data."""

    _code = "finland"
    _name = "Finland"
    _region = "FI"
    _icon = "🇫🇮"
    _resolution = 2
    _settings = FinlandProviderSettings
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _is_base = False
    _extents = [(70.09, 59.45, 31.59, 19.08)]

    _url = "https://avoin-karttakuva.maanmittauslaitos.fi/ortokuvat-ja-korkeusmallit/wcs/v2"
    _wcs_version = "2.0.1"
    _source_crs = "EPSG:3067"
    _tile_size = 1000

    _instructions = (
        "ℹ️ This provider requires an API Key. See [here](https://www.maanmittausl"
        "aitos.fi/rajapinnat/api-avaimen-ohje) for more information on how to create one, then "
        "enter it below in the settings field for API Key."
    )

    def get_wcs_instance_parameters(self):
        if not self.user_settings.api_key:
            raise ValueError("API Key is required for this provider.")

        settings = super().get_wcs_instance_parameters()
        settings["auth"] = Authentication(
            username=self.user_settings.api_key, password=self.user_settings.api_key
        )
        return settings

    def get_wcs_parameters(self, tile: tuple[float, float, float, float]) -> dict:
        return {
            "identifier": ["korkeusmalli_2m"],
            "subsets": [("N", str(tile[0]), str(tile[2])), ("E", str(tile[1]), str(tile[3]))],
            "format": "image/tiff",
            "timeout": 600,
        }



"""This module contains provider of Denmark data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider, DTMProviderSettings


class DenmarkProviderSettings(DTMProviderSettings):
    """Settings for the Denmark provider."""

    token: str = ""


class DenmarkProvider(WCSProvider, DTMProvider):
    """Provider of Denmark data."""

    _code = "denmark"
    _name = "Denmark"
    _region = "DK"
    _icon = "🇩🇰"
    _resolution = 0.4
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _is_base = False
    _settings = DenmarkProviderSettings
    _extents = [(57.7690657013977, 54.4354651516217, 15.5979112056959, 8.00830949937517)]

    _instructions = (
        "ℹ️ This provider requires an access token. See [here](https://confluence"
        ".sdfi.dk/display/MYD/How+to+create+a+user) for more information on "
        "how to create one, then enter it below in the settings field for token."
    )

    _url = "https://api.dataforsyningen.dk/dhm_wcs_DAF"
    _wcs_version = "1.0.0"
    _source_crs = "EPSG:25832"
    _tile_size = 1000

    def get_wcs_parameters(self, tile):
        if not self.user_settings.token:
            raise ValueError("A token is required for this provider.")

        return {
            "identifier": "dhm_terraen",
            "bbox": (tile[1], tile[0], tile[3], tile[2]),
            "crs": "EPSG:25832",
            "width": 2500,
            "height": 2500,
            "format": "GTiff",
            "token": self.user_settings.token,
        }



"""This module contains provider of Canada data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class CanadaProvider(WCSProvider, DTMProvider):
    """Provider of Canada data."""

    _code = "canada"
    _name = "Canada HRDEM"
    _region = "CN"
    _icon = "🇨🇦"
    _resolution = 1
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _is_base = False
    _extents = [(76.49491845750764, 33.66564101989275, -26.69697497450798, -157.7322455868316)]
    _instructions = (
        "HRDEM coverage for Canada is limited. Make sure to check the "
        "[coverage map](https://geo.ca/imagery/high-resolution-digital"
        "-elevation-model-hrdem-canelevation-series/)."
    )

    _url = "https://datacube.services.geo.ca/ows/elevation"
    _wcs_version = "1.1.1"
    _source_crs = "EPSG:3979"
    _tile_size = 1000

    def get_wcs_parameters(self, tile: tuple[float, float, float, float]) -> dict:
        return {
            "identifier": "dtm",
            "gridbasecrs": "urn:ogc:def:crs:EPSG::3979",
            "boundingbox": f"{tile[1]},{tile[0]},{tile[3]},{tile[2]},urn:ogc:def:crs:EPSG::3979",
            "format": "image/geotiff",
            "timeout": 600,
        }



"""This module contains the Config class for map settings and configuration."""

from __future__ import annotations

from maps4fs.generator.component.base.component_xml import XMLComponent


# pylint: disable=R0903
class Config(XMLComponent):
    """Component for map settings and configuration.

    Arguments:
        game (Game): The game instance for which the map is generated.
        coordinates (tuple[float, float]): The latitude and longitude of the center of the map.
        map_size (int): The size of the map in pixels (it's a square).
        map_rotated_size (int): The size of the map in pixels after rotation.
        rotation (int): The rotation angle of the map.
        map_directory (str): The directory where the map files are stored.
        logger (Any, optional): The logger to use. Must have at least three basic methods: debug,
            info, warning. If not provided, default logging will be used.
    """

    def preprocess(self) -> None:
        """Gets the path to the map XML file and saves it to the instance variable."""
        self.xml_path = self.game.map_xml_path(self.map_directory)

    def process(self) -> None:
        """Sets the map size in the map.xml file."""
        self._set_map_size()

    def _set_map_size(self) -> None:
        """Edits map.xml file to set correct map size."""
        tree = self.get_tree()
        if not tree:
            raise FileNotFoundError(f"Map XML file not found: {self.xml_path}")

        root = tree.getroot()
        data = {
            "width": str(self.map_size),
            "height": str(self.map_size),
        }

        for element in root.iter("map"):
            self.update_element(element, data)
            break
        self.save_tree(tree)

    def info_sequence(self) -> dict[str, dict[str, str | float | int]]:
        """Returns information about the component.
        Overview section is needed to create the overview file (in-game map).

        Returns:
            dict[str, dict[str, str | float | int]]: Information about the component.
        """
        # The overview file is exactly 2X bigger than the map size, does not matter
        # if the map is 2048x2048 or 4096x4096, the overview will be 4096x4096
        # and the map will be in the center of the overview.
        # That's why the distance is set to the map height not as a half of it.
        bbox = self.get_bbox(distance=self.map_size)
        south, west, north, east = bbox
        epsg3857_string = self.get_epsg3857_string(bbox=bbox)
        epsg3857_string_with_margin = self.get_epsg3857_string(bbox=bbox, add_margin=True)

        self.qgis_sequence()

        overview_data = {
            "epsg3857_string": epsg3857_string,
            "epsg3857_string_with_margin": epsg3857_string_with_margin,
            "south": south,
            "west": west,
            "north": north,
            "east": east,
            "height": self.map_size * 2,
            "width": self.map_size * 2,
        }

        data = {
            "Overview": overview_data,
        }

        return data  # type: ignore

    def qgis_sequence(self) -> None:
        """Generates QGIS scripts for creating bounding box layers and rasterizing them."""
        bbox = self.get_bbox(distance=self.map_size)
        espg3857_bbox = self.get_espg3857_bbox(bbox=bbox)
        espg3857_bbox_with_margin = self.get_espg3857_bbox(bbox=bbox, add_margin=True)

        qgis_layers = [("Overview_bbox", *espg3857_bbox)]
        qgis_layers_with_margin = [("Overview_bbox_with_margin", *espg3857_bbox_with_margin)]

        layers = qgis_layers + qgis_layers_with_margin

        self.create_qgis_scripts(layers)



"""This module contains provider of Flanders data."""

from maps4fs.generator.dtm.base.wcs import WCSProvider
from maps4fs.generator.dtm.dtm import DTMProvider


class FlandersProvider(WCSProvider, DTMProvider):
    """Provider of Flanders data."""

    _code = "flanders"
    _name = "Flanders DHM II"
    _region = "BE"
    _icon = "🇧🇪"
    _resolution = 1
    _author = "[kbrandwijk](https://github.com/kbrandwijk)"
    _is_community = True
    _is_base = False
    _extents = [(51.5150730375579684, 50.6694992827160817, 5.9444417082210812, 2.5170092434134252)]

    _url = "https://geo.api.vlaanderen.be/el-dtm/wcs"
    _wcs_version = "1.0.0"
    _source_crs = "EPSG:4258"
    _tile_size = 0.02

    def get_wcs_parameters(self, tile: tuple[float, float, float, float]) -> dict:
        return {
            "identifier": "EL.GridCoverage.DTM",
            "bbox": tile,
            "format": "GeoTIFF",
            "crs": "EPSG:4258",
            "width": 1000,
            "height": 1000,
            "timeout": 600,
        }



# pylint: disable=missing-module-docstring
from maps4fs.generator.component.background import Background
from maps4fs.generator.component.base.component import Component
from maps4fs.generator.component.config import Config
from maps4fs.generator.component.grle import GRLE
from maps4fs.generator.component.i3d import I3d
from maps4fs.generator.component.layer import Layer
from maps4fs.generator.component.satellite import Satellite
from maps4fs.generator.component.texture import Texture



"""This module contains the DTMProvider class and its subclasses. DTMProvider class is used to
define different providers of digital terrain models (DTM) data. Each provider has its own URL
and specific settings for downloading and processing the data."""

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Type
from zipfile import ZipFile

import numpy as np
import osmnx as ox
import rasterio
import requests
from pydantic import BaseModel
from rasterio.enums import Resampling
from rasterio.merge import merge
from rasterio.warp import calculate_default_transform, reproject
from tqdm import tqdm

from maps4fs.logger import Logger

if TYPE_CHECKING:
    from maps4fs.generator.map import Map


class DTMProviderSettings(BaseModel):
    """Base class for DTM provider settings models."""


class DTMProvider(ABC):
    """Base class for DTM providers."""

    _code: str | None = None
    _name: str | None = None
    _region: str | None = None
    _icon: str | None = None
    _resolution: float | str | None = None

    _url: str | None = None

    _author: str | None = None
    _contributors: str | None = None
    _is_community: bool = False
    _is_base: bool = False
    _settings: Type[DTMProviderSettings] | None = DTMProviderSettings

    """Bounding box of the provider in the format (north, south, east, west)."""
    _extents: list[tuple[float, float, float, float]] | None = None

    _instructions: str | None = None

    _base_instructions = None

    # AdvancedSettings that should be changed by the DTM Provider if it's selected.
    # * This feature has effect only in the WebUI of the app and ignored in Python package.
    # The first level of the dictionary is a category name,
    # for example: TextureSettings, DEMSettings, etc.
    # The second level is a name of particular setting in the category.
    # Example: {"DEMSettings": {"blur_radius": 35}}
    _default_settings: dict[str, dict[str, Any]] = {}

    def __init__(
        self,
        coordinates: tuple[float, float],
        user_settings: DTMProviderSettings | None,
        size: int,
        directory: str,
        logger: Logger,
        map: Map,
    ):
        self._coordinates = coordinates
        self._user_settings = user_settings
        self._size = size

        if not self._code:
            raise ValueError("Provider code must be defined.")
        self._tile_directory = os.path.join(directory, self._code)
        os.makedirs(self._tile_directory, exist_ok=True)

        self.logger = logger
        self.map = map

    @classmethod
    def default_settings(cls) -> dict[str, dict[str, Any]]:
        """Default settings of the provider.

        Returns:
            dict: Default settings of the provider.
        """
        return cls._default_settings

    @classmethod
    def name(cls) -> str | None:
        """Name of the provider.

        Returns:
            str: Provider name.
        """
        return cls._name

    @classmethod
    def code(cls) -> str | None:
        """Code of the provider.

        Returns:
            str: Provider code.
        """
        return cls._code

    @property
    def coordinates(self) -> tuple[float, float]:
        """Coordinates of the center point of the DTM data.

        Returns:
            tuple[float, float]: Coordinates of the center point of the DTM data.
        """
        return self._coordinates

    @property
    def size(self) -> int:
        """Size of the DTM data in meters.

        Returns:
            int: Size of the DTM data.
        """
        return self._size

    @property
    def url(self) -> str | None:
        """URL of the provider.

        Returns:
            str: URL of the provider or None if not defined.
        """
        return self._url

    def formatted_url(self, **kwargs) -> str:
        """Formatted URL of the provider.

        Arguments:
            **kwargs: Keyword arguments to format the URL.

        Returns:
            str: Formatted URL of the provider.
        """
        if not self.url:
            raise ValueError("URL must be defined.")
        return self.url.format(**kwargs)

    @classmethod
    def author(cls) -> str | None:
        """Author of the provider.

        Returns:
            str: Author of the provider.
        """
        return cls._author

    @classmethod
    def contributors(cls) -> str | None:
        """Contributors of the provider.

        Returns:
            str: Contributors of the provider.
        """
        return cls._contributors

    @classmethod
    def is_base(cls) -> bool:
        """Is the provider a base provider.

        Returns:
            bool: True if the provider is a base provider, False otherwise.
        """
        return cls._is_base

    @classmethod
    def is_community(cls) -> bool:
        """Is the provider a community-driven project.

        Returns:
            bool: True if the provider is a community-driven project, False otherwise.
        """
        return cls._is_community

    @classmethod
    def settings(cls) -> Type[DTMProviderSettings] | None:
        """Settings model of the provider.

        Returns:
            Type[DTMProviderSettings]: Settings model of the provider.
        """
        return cls._settings

    @classmethod
    def instructions(cls) -> str | None:
        """Instructions for using the provider.

        Returns:
            str: Instructions for using the provider.
        """
        return cls._instructions

    @classmethod
    def base_instructions(cls) -> str | None:
        """Instructions for using any provider.

        Returns:
            str: Instructions for using any provider.
        """
        return cls._base_instructions

    @property
    def user_settings(self) -> DTMProviderSettings | None:
        """User settings of the provider.

        Returns:
            DTMProviderSettings: User settings of the provider.
        """
        return self._user_settings

    @classmethod
    def description(cls) -> str:
        """Description of the provider.

        Returns:
            str: Provider description.
        """
        return f"{cls._icon} {cls._region} [{cls._resolution} m/px] {cls._name}"

    @classmethod
    def get_provider_by_code(cls, code: str) -> Type[DTMProvider] | None:
        """Get a provider by its code.

        Arguments:
            code (str): Provider code.

        Returns:
            DTMProvider: Provider class or None if not found.
        """
        for provider in cls.__subclasses__():
            if provider.code() == code:
                return provider
        return None

    @classmethod
    def get_valid_provider_descriptions(
        cls, lat_lon: tuple[float, float], default_code: str = "srtm30"
    ) -> dict[str, str]:
        """Get descriptions of all providers, where keys are provider codes and
        values are provider descriptions.

        Arguments:
            lat_lon (tuple): Latitude and longitude of the center point.
            default_code (str): Default provider code.

        Returns:
            dict: Provider descriptions.
        """
        providers: dict[str, str] = {}
        for provider in cls.__subclasses__():
            if not provider.is_base() and provider.inside_bounding_box(lat_lon):
                code = provider.code()
                if code is not None:
                    providers[code] = provider.description()

        # Sort the dictionary, to make sure that the default provider is the first one.
        providers = dict(sorted(providers.items(), key=lambda item: item[0] != default_code))

        return providers

    @classmethod
    def inside_bounding_box(cls, lat_lon: tuple[float, float]) -> bool:
        """Check if the coordinates are inside the bounding box of the provider.

        Returns:
            bool: True if the coordinates are inside the bounding box, False otherwise.
        """
        lat, lon = lat_lon
        extents = cls._extents
        if extents is None:
            return True
        for extent in extents:
            if extent[0] >= lat >= extent[1] and extent[2] >= lon >= extent[3]:
                return True
        return False

    @abstractmethod
    def download_tiles(self) -> list[str]:
        """Download tiles from the provider.

        Returns:
            list: List of paths to the downloaded tiles.
        """
        raise NotImplementedError

    def get_numpy(self) -> np.ndarray:
        """Get numpy array of the tile.
        Resulting array must be 16 bit (signed or unsigned) integer and it should be already
        windowed to the bounding box of ROI. It also must have only one channel.

        Returns:
            np.ndarray: Numpy array of the tile.
        """
        # download tiles using DTM provider implementation
        tiles = self.download_tiles()
        self.logger.debug(f"Downloaded {len(tiles)} DEM tiles")

        # merge tiles if necessary
        if len(tiles) > 1:
            self.logger.debug("Multiple tiles downloaded. Merging tiles")
            tile, _ = self.merge_geotiff(tiles)
        else:
            tile = tiles[0]

        # determine CRS of the resulting tile and reproject if necessary
        with rasterio.open(tile) as src:
            crs = src.crs
        if crs != "EPSG:4326":
            self.logger.debug(f"Reprojecting GeoTIFF from {crs} to EPSG:4326...")
            tile = self.reproject_geotiff(tile)

        # extract region of interest from the tile
        data = self.extract_roi(tile)

        return data

    # region helpers
    def get_bbox(self) -> tuple[float, float, float, float]:
        """Get bounding box of the tile based on the center point and size.

        Returns:
            tuple: Bounding box of the tile (north, south, east, west).
        """
        west, south, east, north = ox.utils_geo.bbox_from_point(  # type: ignore
            self.coordinates, dist=self.size // 2, project_utm=False
        )
        bbox = float(north), float(south), float(east), float(west)
        return bbox

    def download_tif_files(self, urls: list[str], output_path: str) -> list[str]:
        """Download GeoTIFF files from the given URLs.

        Arguments:
            urls (list): List of URLs to download GeoTIFF files from.
            output_path (str): Path to save the downloaded GeoTIFF files.

        Returns:
            list: List of paths to the downloaded GeoTIFF files.
        """
        tif_files: list[str] = []

        existing_file_urls = [
            f for f in urls if os.path.exists(os.path.join(output_path, os.path.basename(f)))
        ]

        for url in existing_file_urls:
            self.logger.debug("File already exists: %s", os.path.basename(url))
            file_name = os.path.basename(url)
            file_path = os.path.join(output_path, file_name)
            if file_name.endswith(".zip"):
                file_path = self.unzip_img_from_tif(file_name, output_path)
            tif_files.append(file_path)

        for url in tqdm(
            (u for u in urls if u not in existing_file_urls),
            desc="Downloading tiles",
            unit="tile",
            initial=len(tif_files),
            total=len(urls),
        ):
            try:
                file_name = os.path.basename(url)
                file_path = os.path.join(output_path, file_name)
                self.logger.debug("Retrieving TIFF: %s", file_name)

                # Send a GET request to the file URL
                response = requests.get(url, stream=True, timeout=60)
                response.raise_for_status()  # Raise an error for HTTP status codes 4xx/5xx

                # Write the content of the response to the file
                with open(file_path, "wb") as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        file.write(chunk)

                self.logger.debug("File downloaded successfully: %s", file_path)

                if file_name.endswith(".zip"):
                    file_path = self.unzip_img_from_tif(file_name, output_path)

                tif_files.append(file_path)
            except requests.exceptions.RequestException as e:
                self.logger.error("Failed to download file: %s", e)
        return tif_files

    def unzip_img_from_tif(self, file_name: str, output_path: str) -> str:
        """Unpacks the .img file from the zip file.

        Arguments:
            file_name (str): Name of the file to unzip.
            output_path (str): Path to the output directory.

        Returns:
            str: Path to the unzipped file.

        Raises:
            FileNotFoundError: If no .img or .tif file is found in the zip file
        """
        file_path = os.path.join(output_path, file_name)
        img_file_name = file_name.replace(".zip", ".img")
        tif_file_name = file_name.replace(".zip", ".tif")
        img_file_path = os.path.join(output_path, img_file_name)
        tif_file_path = os.path.join(output_path, tif_file_name)
        if os.path.exists(img_file_path):
            self.logger.debug("File already exists: %s", img_file_name)
            return img_file_path
        if os.path.exists(tif_file_path):
            self.logger.debug("File already exists: %s", tif_file_name)
            return tif_file_path
        with ZipFile(file_path, "r") as f_in:
            if img_file_name in f_in.namelist():
                f_in.extract(img_file_name, output_path)
                self.logger.debug("Unzipped file %s to %s", file_name, img_file_name)
                return img_file_path
            if tif_file_name in f_in.namelist():
                f_in.extract(tif_file_name, output_path)
                self.logger.debug("Unzipped file %s to %s", file_name, tif_file_name)
                return tif_file_path
        raise FileNotFoundError("No .img or .tif file found in the zip file.")

    def reproject_geotiff(self, input_tiff: str) -> str:
        """Reproject a GeoTIFF file to a new coordinate reference system (CRS).

        Arguments:
            input_tiff (str): Path to the input GeoTIFF file.

        Returns:
            str: Path to the reprojected GeoTIFF file.
        """
        output_tiff = os.path.join(self._tile_directory, "reprojected.tif")

        # Open the source GeoTIFF
        self.logger.debug("Reprojecting GeoTIFF to EPSG:4326 CRS...")
        with rasterio.open(input_tiff) as src:
            # Get the transform, width, and height of the target CRS
            transform, width, height = calculate_default_transform(
                src.crs, "EPSG:4326", src.width, src.height, *src.bounds
            )

            # Update the metadata for the target GeoTIFF
            kwargs = src.meta.copy()
            kwargs.update(
                {
                    "crs": "EPSG:4326",
                    "transform": transform,
                    "width": width,
                    "height": height,
                    "nodata": None,
                }
            )

            # Open the destination GeoTIFF file and reproject
            with rasterio.open(output_tiff, "w", **kwargs) as dst:
                for i in range(1, src.count + 1):  # Iterate over all raster bands
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=transform,
                        dst_crs="EPSG:4326",
                        resampling=Resampling.average,  # Choose resampling method
                    )

        self.logger.debug("Reprojected GeoTIFF saved to %s", output_tiff)
        return output_tiff

    def merge_geotiff(self, input_files: list[str]) -> tuple[str, str]:
        """Merge multiple GeoTIFF files into a single GeoTIFF file.

        Arguments:
            input_files (list): List of input GeoTIFF files to merge.
        """
        output_file = os.path.join(self._tile_directory, "merged.tif")
        # Open all input GeoTIFF files as datasets
        self.logger.debug("Merging tiff files...")
        datasets = [rasterio.open(file) for file in input_files]

        # Merge datasets
        crs = datasets[0].crs
        mosaic, out_transform = merge(datasets, nodata=0)

        # Get metadata from the first file and update it for the output
        out_meta = datasets[0].meta.copy()
        out_meta.update(
            {
                "driver": "GTiff",
                "height": mosaic.shape[1],
                "width": mosaic.shape[2],
                "transform": out_transform,
                "count": mosaic.shape[0],  # Number of bands
            }
        )

        # Write merged GeoTIFF to the output file
        with rasterio.open(output_file, "w", **out_meta) as dest:
            dest.write(mosaic)

        self.logger.debug("GeoTIFF images merged successfully into %s", output_file)
        return output_file, crs

    def extract_roi(self, tile_path: str) -> np.ndarray:
        """Extract region of interest (ROI) from the GeoTIFF file.

        Arguments:
            tile_path (str): Path to the GeoTIFF file.

        Raises:
            ValueError: If the tile does not contain any data.

        Returns:
            np.ndarray: Numpy array of the ROI.
        """
        north, south, east, west = self.get_bbox()
        with rasterio.open(tile_path) as src:
            self.logger.debug("Opened tile, shape: %s, dtype: %s.", src.shape, src.dtypes[0])
            window = rasterio.windows.from_bounds(west, south, east, north, src.transform)
            self.logger.debug(
                "Window parameters. Column offset: %s, row offset: %s, width: %s, height: %s.",
                window.col_off,
                window.row_off,
                window.width,
                window.height,
            )
            data = src.read(1, window=window, masked=True)
        if not data.size > 0:
            raise ValueError("No data in the tile.")

        return data

    # endregion



"""This module contains provider of Thuringia data."""

import os
from maps4fs.generator.dtm import utils
from maps4fs.generator.dtm.dtm import DTMProvider

class ThuringiaProvider(DTMProvider):
    """Provider of Thuringia data.
    Data is provided by the Kompetenzzentrum Geodateninfrastruktur Thüringen (© GDI-Th) and available
    at https://geoportal.thueringen.de/gdi-th/download-offene-geodaten/download-hoehendaten under BY 2.0 license.
    """
    _code = "thuringia"
    _name = "Thüringen DGM1"
    _region = "DE"
    _icon = "🇩🇪󠁥󠁢󠁹󠁿"
    _resolution = 1
    _author = "[H4rdB4se](https://github.com/H4rdB4se)"
    _is_community = True
    _instructions = None
    _extents = [(51.5997, 50.2070, 12.69674, 9.8548)]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tiff_path = os.path.join(self._tile_directory, "tiffs")
        os.makedirs(self.tiff_path, exist_ok=True)

    def download_tiles(self) -> list[str]:
        bbox = self.get_bbox()
        west, east, north, south = utils.transform_bbox(bbox, "EPSG:25832")
        download_urls = self.get_download_urls(north, south, east, west)
        all_tif_files = self.download_tif_files(download_urls, self.tiff_path)
        return all_tif_files

    @staticmethod
    def get_first_n_digits(value: float, digits: int) -> int:
        """Get the first n digits of a number."""
        return int(str(value)[:digits])

    def get_download_urls(self, north: float, south: float, east: float, west: float) -> list[str]:
        """Calculate all possible tiles within the bounding box.

        Args:
            north (float): Northern boundary.
            south (float): Southern boundary.
            east (float): Eastern boundary.
            west (float): Western boundary.

        Returns:
            list: List of tile names.
        """
        urls = []
        lat = self.get_first_n_digits(south, 4)
        while lat <= self.get_first_n_digits(north, 4):
            lon = self.get_first_n_digits(west, 3)
            while lon <= self.get_first_n_digits(east, 3):
                tile_url = f"https://geoportal.geoportal-th.de/hoehendaten/DGM/dgm_2020-2025/dgm1_32_{lon}_{lat}_1_th_2020-2025.zip"
                urls.append(tile_url)
                lon += 1
            lat += 1
        return urls



"""Utility functions for the DTM generator."""

import numpy as np
from pyproj import Transformer


def tile_bbox(
    bbox: tuple[float, float, float, float], tile_size: float
) -> list[tuple[float, float, float, float]]:
    """Tile the bounding box into smaller bounding boxes of a specified size.

    Arguments:
        bbox (tuple): Bounding box to tile (north, south, east, west).
        tile_size (int): Size of the tiles in meters.

    Returns:
        list: List of smaller bounding boxes (north, south, east, west).
    """
    north, south, east, west = bbox
    x_coords = np.arange(west, east, tile_size if east > west else -tile_size)
    y_coords = np.arange(north, south, tile_size if south > north else -tile_size)
    x_coords = np.append(x_coords, east).astype(x_coords.dtype)
    y_coords = np.append(y_coords, south).astype(y_coords.dtype)

    x_min, y_min = np.meshgrid(x_coords[:-1], y_coords[:-1], indexing="ij")
    x_max, y_max = np.meshgrid(x_coords[1:], y_coords[1:], indexing="ij")

    tiles = np.stack(
        [x_min.ravel(), y_min.ravel(), x_max.ravel(), y_max.ravel()], axis=1, dtype=float
    )

    return [tuple(tile[i].item() for i in range(4)) for tile in tiles]


def transform_bbox(
    bbox: tuple[float, float, float, float], to_crs: str
) -> tuple[float, float, float, float]:
    """Transform the bounding box to a different coordinate reference system (CRS).

    Arguments:
        bbox (tuple): Bounding box to transform (north, south, east, west).
        to_crs (str): Target CRS (e.g., EPSG:4326 for CRS:84).

    Returns:
        tuple: Transformed bounding box (north, south, east, west).
    """
    transformer = Transformer.from_crs("epsg:4326", to_crs)
    north, south, east, west = bbox

    # EPSG:4326 is lat, lon, so xx is lat and yy is lon
    bottom_left_x, bottom_left_y = transformer.transform(xx=south, yy=west)
    top_left_x, top_left_y = transformer.transform(xx=north, yy=west)
    top_right_x, top_right_y = transformer.transform(xx=north, yy=east)
    bottom_right_x, bottom_right_y = transformer.transform(xx=south, yy=east)

    west = min(bottom_left_y, bottom_right_y)
    east = max(top_left_y, top_right_y)
    north = min(bottom_left_x, top_left_x)
    south = max(bottom_right_x, top_right_x)

    return north, south, east, west



"""This module contains the base class for all map generation components."""

from __future__ import annotations

import json
import os
from copy import deepcopy
from typing import TYPE_CHECKING, Any

import cv2
import osmnx as ox
from pyproj import Transformer
from shapely.affinity import rotate, translate
from shapely.geometry import LineString, Polygon, box

from maps4fs.generator.qgis import save_scripts

if TYPE_CHECKING:
    from maps4fs.generator.game import Game
    from maps4fs.generator.map import Map


class Component:
    """Base class for all map generation components.

    Arguments:
        game (Game): The game instance for which the map is generated.
        map (Map): The map instance for which the component is generated.
        coordinates (tuple[float, float]): The latitude and longitude of the center of the map.
        map_size (int): The size of the map in pixels.
        map_rotated_size (int): The size of the map in pixels after rotation.
        rotation (int): The rotation angle of the map.
        map_directory (str): The directory where the map files are stored.
        logger (Any, optional): The logger to use. Must have at least three basic methods: debug,
            info, warning. If not provided, default logging will be used.
    """

    def __init__(
        self,
        game: Game,
        map: Map,
        coordinates: tuple[float, float],
        map_size: int,
        map_rotated_size: int,
        rotation: int,
        map_directory: str,
        logger: Any = None,
        **kwargs: dict[str, Any],
    ):
        self.game = game
        self.map = map
        self.coordinates = coordinates
        self.map_size = map_size
        self.map_rotated_size = map_rotated_size
        self.rotation = rotation
        self.map_directory = map_directory
        self.logger = logger
        self.kwargs = kwargs

        self.logger.debug(
            "Component %s initialized. Map size: %s, map rotated size: %s",
            self.__class__.__name__,
            self.map_size,
            self.map_rotated_size,
        )

        os.makedirs(self.previews_directory, exist_ok=True)
        os.makedirs(self.scripts_directory, exist_ok=True)
        os.makedirs(self.info_layers_directory, exist_ok=True)
        os.makedirs(self.satellite_directory, exist_ok=True)

        self.save_bbox()
        self.preprocess()

    def preprocess(self) -> None:
        """Prepares the component for processing. Must be implemented in the child class."""
        return

    def process(self) -> None:
        """Launches the component processing. Must be implemented in the child class.

        Raises:
            NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError

    def previews(self) -> list[str]:
        """Returns a list of paths to the preview images. If the component does not generate any
        previews, the method may not be re-implemented in the child class.

        Returns:
            list[str]: A list of paths to the preview images.
        """
        return []

    @property
    def previews_directory(self) -> str:
        """The directory where the preview images are stored.

        Returns:
            str: The directory where the preview images are stored.
        """
        return os.path.join(self.map_directory, "previews")

    @property
    def info_layers_directory(self) -> str:
        """The directory where the info layers are stored.

        Returns:
            str: The directory where the info layers are stored.
        """
        return os.path.join(self.map_directory, "info_layers")

    @property
    def scripts_directory(self) -> str:
        """The directory where the scripts are stored.

        Returns:
            str: The directory where the scripts are stored.
        """
        return os.path.join(self.map_directory, "scripts")

    @property
    def satellite_directory(self) -> str:
        """The directory where the satellite images are stored.

        Returns:
            str: The directory where the satellite images are stored.
        """
        return os.path.join(self.map_directory, "satellite")

    @property
    def generation_info_path(self) -> str:
        """The path to the generation info JSON file.

        Returns:
            str: The path to the generation info JSON file.
        """
        return os.path.join(self.map_directory, "generation_info.json")

    def info_sequence(self) -> dict[Any, Any]:
        """Returns the information sequence for the component. Must be implemented in the child
        class. If the component does not have an information sequence, an empty dictionary must be
        returned.

        Returns:
            dict[Any, Any]: The information sequence for the component.
        """
        return {}

    def commit_generation_info(self) -> None:
        """Commits the generation info to the generation info JSON file."""
        self.update_generation_info(self.info_sequence())

    def update_generation_info(self, data: dict[Any, Any]) -> None:
        """Updates the generation info with the provided data.
        If the generation info file does not exist, it will be created.

        Arguments:
            data (dict[Any, Any]): The data to update the generation info with.
        """
        if os.path.isfile(self.generation_info_path):
            with open(self.generation_info_path, "r", encoding="utf-8") as file:
                generation_info = json.load(file)
                self.logger.debug("Loaded generation info from %s", self.generation_info_path)
        else:
            self.logger.debug(
                "Generation info file does not exist, creating a new one in %s",
                self.generation_info_path,
            )
            generation_info = {}

        updated_generation_info = deepcopy(generation_info)
        updated_generation_info[self.__class__.__name__] = data

        self.logger.debug("Updated generation info, now contains %s fields", len(data))

        with open(self.generation_info_path, "w", encoding="utf-8") as file:
            try:
                json.dump(updated_generation_info, file, indent=4)
            except Exception as e:
                self.logger.warning("Could not save updated generation info: %s", e)

        self.logger.debug("Saved updated generation info to %s", self.generation_info_path)

    def get_bbox(
        self,
        coordinates: tuple[float, float] | None = None,
        distance: int | None = None,
    ) -> tuple[float, float, float, float]:
        """Calculates the bounding box of the map from the coordinates and the height and
        width of the map.
        If coordinates and distance are not provided, the instance variables are used.

        Arguments:
            coordinates (tuple[float, float], optional): The latitude and longitude of the center
                of the map. Defaults to None.
            distance (int, optional): The distance from the center of the map to the edge of the
                map in all directions. Defaults to None.

        Returns:
            tuple[float, float, float, float]: The bounding box of the map.
        """
        coordinates = coordinates or self.coordinates
        distance = distance or int(self.map_rotated_size / 2)

        west, south, east, north = ox.utils_geo.bbox_from_point(
            coordinates,
            dist=distance,
        )

        bbox = north, south, east, west
        self.logger.debug(
            "Calculated bounding box for component: %s: %s, distance: %s",
            self.__class__.__name__,
            bbox,
            distance,
        )
        return bbox

    def save_bbox(self) -> None:
        """Saves the bounding box of the map to the component instance from the coordinates and the
        height and width of the map.
        """
        self.bbox = self.get_bbox()
        self.logger.debug("Saved bounding box: %s", self.bbox)

    @property
    def new_bbox(self) -> tuple[float, float, float, float]:
        """This property is used for a new version of osmnx library, where the order of coordinates
        has been changed to (left, bottom, right, top).

        Returns:
            tuple[float, float, float, float]: The bounding box of the map in the new order:
                (left, bottom, right, top).
        """
        # FutureWarning: The expected order of coordinates in `bbox`
        # will change in the v2.0.0 release to `(left, bottom, right, top)`.
        north, south, east, west = self.bbox
        return west, south, east, north

    def get_espg3857_bbox(
        self, bbox: tuple[float, float, float, float] | None = None, add_margin: bool = False
    ) -> tuple[float, float, float, float]:
        """Converts the bounding box to EPSG:3857.
        If the bounding box is not provided, the instance variable is used.

        Arguments:
            bbox (tuple[float, float, float, float], optional): The bounding box to convert.
            add_margin (bool, optional): Whether to add a margin to the bounding box.

        Returns:
            tuple[float, float, float, float]: The bounding box in EPSG:3857.
        """
        bbox = bbox or self.bbox
        north, south, east, west = bbox
        transformer = Transformer.from_crs("epsg:4326", "epsg:3857")
        epsg3857_north, epsg3857_west = transformer.transform(north, west)
        epsg3857_south, epsg3857_east = transformer.transform(south, east)

        if add_margin:
            margin = 500
            epsg3857_north = int(epsg3857_north - margin)
            epsg3857_south = int(epsg3857_south + margin)
            epsg3857_east = int(epsg3857_east - margin)
            epsg3857_west = int(epsg3857_west + margin)

        return epsg3857_north, epsg3857_south, epsg3857_east, epsg3857_west

    def get_epsg3857_string(
        self, bbox: tuple[float, float, float, float] | None = None, add_margin: bool = False
    ) -> str:
        """Converts the bounding box to EPSG:3857 string.
        If the bounding box is not provided, the instance variable is used.

        Arguments:
            bbox (tuple[float, float, float, float], optional): The bounding box to convert.
            add_margin (bool, optional): Whether to add a margin to the bounding box.

        Returns:
            str: The bounding box in EPSG:3857 string.
        """
        north, south, east, west = self.get_espg3857_bbox(bbox, add_margin=add_margin)
        return f"{north},{south},{east},{west} [EPSG:3857]"

    def create_qgis_scripts(
        self, qgis_layers: list[tuple[str, float, float, float, float]]
    ) -> None:
        """Creates QGIS scripts from the given layers.
        Each layer is a tuple where the first element is a name of the layer and the rest are the
        bounding box coordinates in EPSG:3857.
        For filenames, the class name is used as a prefix.

        Arguments:
            qgis_layers (list[tuple[str, float, float, float, float]]): The list of layers to
                create scripts for.
        """
        class_name = self.__class__.__name__.lower()
        save_scripts(qgis_layers, class_name, self.scripts_directory)

    def get_polygon_center(self, polygon_points: list[tuple[int, int]]) -> tuple[int, int]:
        """Calculates the center of a polygon defined by a list of points.

        Arguments:
            polygon_points (list[tuple[int, int]]): The points of the polygon.

        Returns:
            tuple[int, int]: The center of the polygon.
        """
        polygon = Polygon(polygon_points)
        center = polygon.centroid
        return int(center.x), int(center.y)

    def absolute_to_relative(
        self, point: tuple[int, int], center: tuple[int, int]
    ) -> tuple[int, int]:
        """Converts a pair of absolute coordinates to relative coordinates.

        Arguments:
            point (tuple[int, int]): The absolute coordinates.
            center (tuple[int, int]): The center coordinates.

        Returns:
            tuple[int, int]: The relative coordinates.
        """
        cx, cy = center
        x, y = point
        return x - cx, y - cy

    def top_left_coordinates_to_center(self, top_left: tuple[int, int]) -> tuple[int, int]:
        """Converts a pair of coordinates from the top-left system to the center system.
        In top-left system, the origin (0, 0) is in the top-left corner of the map, while in the
        center system, the origin is in the center of the map.

        Arguments:
            top_left (tuple[int, int]): The coordinates in the top-left system.

        Returns:
            tuple[int, int]: The coordinates in the center system.
        """
        x, y = top_left
        cs_x = x - self.map_size // 2
        cs_y = y - self.map_size // 2

        return cs_x, cs_y

    def fit_object_into_bounds(
        self,
        polygon_points: list[tuple[int, int]] | None = None,
        linestring_points: list[tuple[int, int]] | None = None,
        margin: int = 0,
        angle: int = 0,
        border: int = 0,
    ) -> list[tuple[int, int]]:
        """Fits a polygon into the bounds of the map.

        Arguments:
            polygon_points (list[tuple[int, int]]): The points of the polygon.
            linestring_points (list[tuple[int, int]]): The points of the linestring.
            margin (int, optional): The margin to add to the polygon. Defaults to 0.
            angle (int, optional): The angle to rotate the polygon by. Defaults to 0.
            border (int, optional): The border to add to the bounds. Defaults to 0.

        Returns:
            list[tuple[int, int]]: The points of the polygon fitted into the map bounds.
        """
        if polygon_points is None and linestring_points is None:
            raise ValueError("Either polygon or linestring points must be provided.")

        min_x = min_y = 0 + border
        max_x = max_y = self.map_size - border

        object_type = Polygon if polygon_points else LineString

        # polygon = Polygon(polygon_points)
        osm_object = object_type(polygon_points or linestring_points)

        if angle:
            center_x = center_y = self.map_rotated_size // 2
            self.logger.debug(
                "Rotating the osm_object by %s degrees with center at %sx%s",
                angle,
                center_x,
                center_y,
            )
            osm_object = rotate(osm_object, -angle, origin=(center_x, center_y))
            offset = (self.map_size / 2) - (self.map_rotated_size / 2)
            self.logger.debug("Translating the osm_object by %s", offset)
            osm_object = translate(osm_object, xoff=offset, yoff=offset)
            self.logger.debug("Rotated and translated the osm_object.")

        if margin and object_type is Polygon:
            osm_object = osm_object.buffer(margin, join_style="mitre")
            if osm_object.is_empty:
                raise ValueError("The osm_object is empty after adding the margin.")

        # Create a bounding box for the map bounds
        bounds = box(min_x, min_y, max_x, max_y)

        # Intersect the osm_object with the bounds to fit it within the map
        try:
            fitted_osm_object = osm_object.intersection(bounds)
            self.logger.debug("Fitted the osm_object into the bounds: %s", bounds)
        except Exception as e:
            raise ValueError(f"Could not fit the osm_object into the bounds: {e}")

        if not isinstance(fitted_osm_object, object_type):
            raise ValueError("The fitted osm_object is not valid (probably splitted into parts).")

        # Return the fitted polygon points
        if object_type is Polygon:
            as_list = list(fitted_osm_object.exterior.coords)
        elif object_type is LineString:
            as_list = list(fitted_osm_object.coords)
        else:
            raise ValueError("The object type is not supported.")

        if not as_list:
            raise ValueError("The fitted osm_object has no points.")
        return as_list

    def get_infolayer_path(self, layer_name: str) -> str | None:
        """Returns the path to the info layer file.

        Arguments:
            layer_name (str): The name of the layer.

        Returns:
            str | None: The path to the info layer file or None if the layer does not exist.
        """
        info_layer_path = os.path.join(self.info_layers_directory, f"{layer_name}.json")
        if not os.path.isfile(info_layer_path):
            self.logger.warning("Info layer %s does not exist", info_layer_path)
            return None
        return info_layer_path

    def get_infolayer_data(self, layer_name: str, layer_key: str) -> Any | None:
        """Reads the JSON file of the requested info layer and returns the value of the requested
        key. If the layer or the key does not exist, None is returned.

        Arguments:
            layer_name (str): The name of the layer.
            layer_key (str): The key to get the value of.

        Returns:
            Any | None: The value of the requested key or None if the layer or the key does not
                exist.
        """
        infolayer_path = self.get_infolayer_path(layer_name)
        if not infolayer_path:
            return None

        with open(infolayer_path, "r", encoding="utf-8") as file:
            data = json.load(file)

        return data.get(layer_key)

    def rotate_image(
        self,
        image_path: str,
        angle: int,
        output_height: int,
        output_width: int,
        output_path: str | None = None,
    ) -> None:
        """Rotates an image by a given angle around its center and cuts out the center to match
        the output size.

        Arguments:
            image_path (str): The path to the image to rotate.
            angle (int): The angle to rotate the image by.
            output_height (int): The height of the output image.
            output_width (int): The width of the output image.
        """
        if not os.path.isfile(image_path):
            self.logger.warning("Image %s does not exist", image_path)
            return

        image = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
        if image is None:
            self.logger.warning("Image %s could not be read", image_path)
            return

        self.logger.debug("Read image from %s with shape: %s", image_path, image.shape)

        if not output_path:
            output_path = image_path

        height, width = image.shape[:2]
        center = (width // 2, height // 2)

        self.logger.debug(
            "Rotating the image... Angle: %s, center: %s, height: %s, width: %s",
            angle,
            center,
            height,
            width,
        )

        rotation_matrix = cv2.getRotationMatrix2D(center, angle, 1.0)
        rotated = cv2.warpAffine(image, rotation_matrix, (width, height))

        start_x = center[0] - output_width // 2
        start_y = center[1] - output_height // 2
        end_x = start_x + output_width
        end_y = start_y + output_height

        self.logger.debug(
            "Cropping the rotated image: start_x: %s, start_y: %s, end_x: %s, end_y: %s",
            start_x,
            start_y,
            end_x,
            end_y,
        )

        cropped = rotated[start_y:end_y, start_x:end_x]

        self.logger.debug("Shape of the cropped image: %s", cropped.shape)

        cv2.imwrite(output_path, cropped)

    @staticmethod
    def interpolate_points(
        polyline: list[tuple[int, int]], num_points: int = 4
    ) -> list[tuple[int, int]]:
        """Receives a list of tuples, which represents a polyline. Add additional points
        between the existing points to make the polyline smoother.

        Arguments:
            polyline (list[tuple[int, int]]): The list of points to interpolate.
            num_points (int): The number of additional points to add between each pair of points.

        Returns:
            list[tuple[int, int]]: The list of points with additional points.
        """
        if not polyline or num_points < 1:
            return polyline

        interpolated_polyline = []
        for i in range(len(polyline) - 1):
            p1 = polyline[i]
            p2 = polyline[i + 1]
            interpolated_polyline.append(p1)
            for j in range(1, num_points + 1):
                new_point = (
                    p1[0] + (p2[0] - p1[0]) * j / (num_points + 1),
                    p1[1] + (p2[1] - p1[1]) * j / (num_points + 1),
                )
                interpolated_polyline.append((int(new_point[0]), int(new_point[1])))
        interpolated_polyline.append(polyline[-1])

        return interpolated_polyline

    def get_z_scaling_factor(self, ignore_height_scale_multiplier: bool = False) -> float:
        """Calculates the scaling factor for the Z axis based on the map settings.

        Returns:
            float -- The scaling factor for the Z axis.
        """

        scaling_factor = 1 / self.map.dem_settings.multiplier

        if self.map.shared_settings.height_scale_multiplier and not ignore_height_scale_multiplier:
            scaling_factor *= self.map.shared_settings.height_scale_multiplier
        if self.map.shared_settings.mesh_z_scaling_factor:
            scaling_factor *= 1 / self.map.shared_settings.mesh_z_scaling_factor

        return scaling_factor
